from cython cimport freelist
from brainpy._c.composition cimport Composition, ElementHashTable, Element
from brainpy._c.isotopic_constants cimport IsotopicConstants
from brainpy._c.double_vector cimport DoubleVector


cdef struct Peak:
    double mz
    double intensity
    int charge


cdef struct PeakList:
    Peak* peaks
    size_t used
    size_t size


cdef Peak* make_peak(double mz, double intensity, int charge) nogil
cdef void print_peak(Peak* peak) nogil

cdef PeakList* make_peak_list() nogil
cdef void free_peak_list(PeakList* peaklist) nogil
cdef int resize_peak_list(PeakList* peaklist) nogil
cdef void peak_list_append(PeakList* peaklist, Peak* peak) nogil
cdef void peak_list_reset(PeakList* peaklist) nogil

cdef PeakList* peak_list_ignore_below(PeakList* peaklist, double ignore_below, PeakList* result) nogil
cdef PeakList* peak_list_truncate_after(PeakList* peaklist, double truncate_after, PeakList* result) nogil
cdef void peak_list_shift(PeakList* peaklist, double shift) nogil

cdef list peaklist_to_list(PeakList* peaklist)


cdef struct ElementCache:
    Element** elements
    size_t used
    size_t size
    ElementHashTable* source


cdef ElementCache* make_element_cache(ElementHashTable* source) nogil
cdef void free_element_cache(ElementCache* cache) nogil
cdef int resize_element_cache(ElementCache* cache) nogil
cdef int element_cache_put(ElementCache* cache, Element** element) nogil
cdef int element_cache_get(ElementCache* cache, char* symbol, Element** out) nogil


cdef struct IsotopicDistribution:
    Composition* composition
    IsotopicConstants* _isotopic_constants
    int order
    double average_mass
    Peak* monoisotopic_peak
    ElementCache* cache


cdef IsotopicDistribution* make_isotopic_distribution(Composition* composition, int order, ElementCache* cache=*) nogil
cdef void free_isotopic_distribution(IsotopicDistribution* distribution)  nogil


@freelist(100000)
cdef class TheoreticalPeak(object):
    cdef:
        public double mz
        public double intensity
        public int charge

    @staticmethod
    cdef TheoreticalPeak _create(double mz, double intensity, int charge)

    cpdef bint _eq(self, TheoreticalPeak other)
    cpdef TheoreticalPeak clone(self)


cdef int max_variants(Composition* composition, ElementCache* cache) nogil
cdef int guess_npeaks(Composition* composition_struct, size_t max_npeaks, ElementCache* cache=*) nogil

cpdef list _isotopic_variants(object composition, object npeaks=*, int charge=*, double charge_carrier=*)


cdef PeakList* isotopic_variants(Composition* composition, int npeaks, int charge=*, double charge_carrier=*) nogil



import unittest

from brainpy.composition import _parse_formula as parse_formula, calculate_mass
from brainpy import _has_c

if _has_c:
    from brainpy._c.composition import parse_formula as cparse_formula


class CompositionTest(unittest.TestCase):
    def test_parse(self):
        formula = "C6H12O6"
        composition = parse_formula(formula)
        self.assertEqual(composition, {"C": 6, "H": 12, "O": 6})
        self.assertEqual(composition * 2, {"C": 6 * 2, "H": 12 * 2, "O": 6 * 2})
        if _has_c:
            composition = cparse_formula(formula)
            self.assertEqual(composition, {"C": 6, "H": 12, "O": 6})
            self.assertEqual(composition * 2, {"C": 6 * 2, "H": 12 * 2, "O": 6 * 2})

    def test_isotope_parse(self):
        for formula in ["O1H1H1H[2]1", "O1H1H[2]1H1"]:
            composition = parse_formula(formula)
            self.assertEqual(composition, {"O": 1, "H": 2, "H[2]": 1}, msg="%r != %r" % (
                composition, {"O": 1, "H": 2, "H[2]": 1}))
            self.assertAlmostEqual(composition.mass(), 20.0246, 3, msg="%r.mass() (%r) != %r" % (
                composition, composition.mass(), 20.0246))
            self.assertEqual(composition['H'], 2)
            if _has_c:
                composition = cparse_formula(formula)
                self.assertEqual(composition, {"O": 1, "H": 2, "H[2]": 1}, msg="%r != %r" % (
                    composition, {"O": 1, "H": 2, "H[2]": 1}))
                self.assertAlmostEqual(composition.mass(), 20.0246, 3, msg="%r.mass() (%r) != %r" % (
                    composition, composition.mass(), 20.0246))
                self.assertEqual(composition['H'], 2)

    def test_mass(self):
        formula = "C6H12O6"
        composition = parse_formula(formula)
        self.assertAlmostEqual(composition.mass(), calculate_mass({"C": 6, "H": 12, "O": 6}))
        if _has_c:
            composition = cparse_formula(formula)
            self.assertAlmostEqual(composition.mass(), calculate_mass({"C": 6, "H": 12, "O": 6}))



from brainpy.mass_dict import nist_mass as __nist_mass

ctypedef int count_type

cdef dict nist_mass
nist_mass = __nist_mass

cdef double PROTON = nist_mass["H+"][0][0]

cdef double neutral_mass(double mz,  int z, double charge_carrier=*) nogil
cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*) nogil
cdef char* _parse_isotope_string(char* label, int* isotope_num, char* element_name) nogil

# -----------------------------------------------------------------------------
# Isotope and IsotopeMap Declarations

cdef struct Isotope:
    double mass
    double abundance
    int neutrons
    int neutron_shift

cdef struct IsotopeMap:
    Isotope* bins
    size_t size

cdef IsotopeMap* make_isotope_map(list organized_isotope_data, size_t size)

cdef Isotope* get_isotope_by_neutron_shift(IsotopeMap* isotopes, int neutron_shift) nogil
cdef void free_isotope_map(IsotopeMap* isotopes) nogil

# -----------------------------------------------------------------------------
# Element Declarations

cdef struct Element:
    char* symbol
    IsotopeMap* isotopes
    int monoisotopic_isotope_index

cdef void _isotopes_of(char* element_symbol, IsotopeMap** isotope_frequencies)
cdef Element* make_element(char* symbol)

cdef double element_monoisotopic_mass(Element* element) nogil
cdef int element_min_neutron_shift(Element* element) nogil
cdef int element_max_neutron_shift(Element* element) nogil
cdef void free_element(Element* element) nogil

cdef Element* make_fixed_isotope_element(Element* element, int neutron_count) nogil


# -----------------------------------------------------------------------------
# ElementHashTable and ElementHashBucket Declarations

cdef struct ElementHashBucket:
    Element** elements
    size_t used
    size_t size


cdef void free_element_hash_bucket(ElementHashBucket* bucket) nogil


cdef struct ElementHashTable:
    ElementHashBucket* buckets
    size_t size

cdef ElementHashTable* _ElementTable

cdef ElementHashTable* make_element_hash_table(size_t size) nogil

cdef int element_hash_bucket_insert(ElementHashBucket* bucket, Element* element) nogil

cdef int element_hash_bucket_find(ElementHashBucket* bucket, char* symbol, Element** out) nogil

cdef int element_hash_table_get(ElementHashTable* table, char* symbol, Element** out) nogil

cdef int element_hash_table_put(ElementHashTable* table, Element* element) nogil

cdef size_t hash_string(char *str) nogil

cdef size_t free_element_hash_table(ElementHashTable* table) nogil


cdef ElementHashTable* get_system_element_hash_table() nogil

cdef int set_system_element_hash_table(ElementHashTable* table) nogil


# -----------------------------------------------------------------------------
# Composition Declarations

cdef struct Composition:
    char** elements
    count_type* counts
    size_t size
    size_t used
    int max_variants

cdef Composition* make_composition() nogil
cdef Composition* copy_composition(Composition* composition) nogil
cdef void print_composition(Composition* composition) nogil
cdef int composition_set_element_count(Composition* composition, char* element, count_type count) nogil
cdef int composition_get_element_count(Composition* composition, char* element, count_type* count) nogil
cdef int composition_inc_element_count(Composition* composition, char* element, count_type increment) nogil
cdef int composition_resize(Composition* composition) nogil
cdef double composition_mass(Composition* composition) nogil
cdef void free_composition(Composition* composition) nogil

cdef Composition* composition_add(Composition* composition_1, Composition* composition_2, int sign) nogil
cdef int composition_iadd(Composition* composition_1, Composition* composition_2, int sign) nogil
cdef Composition* composition_mul(Composition* composition, int scale) nogil
cdef void composition_imul(Composition* composition, int scale) nogil

cdef dict composition_to_dict(Composition* composition)
cdef Composition* dict_to_composition(dict comp_dict)

cdef class PyComposition(object):
    cdef:
        Composition* impl
        public double cached_mass
        public bint _clean
    @staticmethod
    cdef PyComposition _create(Composition* base)
    cdef void _set_impl(self, Composition* composition, bint free_existing=*)

    cpdef double mass(self)
    cpdef bint __equality_pycomposition(self, PyComposition other)
    cpdef bint __equality_dict(self, dict other)
    cpdef PyComposition copy(self)

cpdef PyComposition parse_formula(str formula)



# cython: embedsignature=True
# cython: profile=True

from cpython.list cimport PyList_GET_ITEM, PyList_GET_SIZE, PyList_Append
from cpython.int cimport PyInt_FromLong
from cpython.float cimport PyFloat_FromDouble, PyFloat_AsDouble
from cpython.dict cimport PyDict_GetItem, PyDict_SetItem, PyDict_Keys, PyDict_Next
from cpython.object cimport PyObject

from libc.math cimport log, exp

import operator
import re

mz_getter = operator.attrgetter("mz")

from mass_dict import nist_mass as _nist_mass
from brainpy._c.isotopic_distribution cimport TheoreticalPeak as Peak


nist_mass = _nist_mass
PROTON = nist_mass["H+"][0][0]


cdef double neutral_mass(double mz,  int z, double charge_carrier=PROTON):
    return (mz * abs(z)) - (z * charge_carrier)


cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=PROTON):
    return (neutral_mass + (z * charge_carrier)) / abs(z)


cdef void _update_elementary_symmetric_polynomial(list power_sum, list elementary_symmetric_polynomial, size_t order):
    cdef:
        size_t begin, end, k, j
        double el
        int sign
    begin = PyList_GET_SIZE(elementary_symmetric_polynomial)
    end = PyList_GET_SIZE(power_sum)
    for k in range(begin, end):
        if k == 0:
            PyList_Append(elementary_symmetric_polynomial, 1.0)
        elif k > order:
            PyList_Append(elementary_symmetric_polynomial, 0.)
        else:
            el = 0.
            for j in range(1, k + 1):
                sign = 1 if (j % 2) == 1 else -1
                el += sign * PyFloat_AsDouble(<object>PyList_GET_ITEM(power_sum, j)) * PyFloat_AsDouble(<object>PyList_GET_ITEM(elementary_symmetric_polynomial, k - j))
            el /= <double>(k)
            PyList_Append(elementary_symmetric_polynomial, el)

cdef void _update_power_sum(list ps_vec, list esp_vec, size_t order):
    cdef:
        size_t begin, end, k, j
        int sign
        double temp_ps
    begin = PyList_GET_SIZE(ps_vec)
    end = PyList_GET_SIZE(esp_vec)
    for k in range(begin, end):
        if k == 0:
            PyList_Append(ps_vec, 0.)
            continue
        temp_ps = 0.
        sign = -1
        for j in range(1, k):
            sign *= -1
            temp_ps += sign * PyFloat_AsDouble(<object>PyList_GET_ITEM(esp_vec, j)) * PyFloat_AsDouble(<object>PyList_GET_ITEM(ps_vec, k - j))
        sign *= -1
        temp_ps += sign * PyFloat_AsDouble(<object>PyList_GET_ITEM(esp_vec, k)) * k
        PyList_Append(ps_vec, temp_ps)

cdef void newton(list power_sum, list elementary_symmetric_polynomial, int order):
    if PyList_GET_SIZE(power_sum) > PyList_GET_SIZE(elementary_symmetric_polynomial):
        _update_elementary_symmetric_polynomial(power_sum, elementary_symmetric_polynomial, order)
    elif PyList_GET_SIZE(power_sum) < PyList_GET_SIZE(elementary_symmetric_polynomial):
        _update_power_sum(power_sum, elementary_symmetric_polynomial, order)

cdef list vietes(list coefficients):
    cdef:
        list elementary_symmetric_polynomial
        double tail, el
        size_t size, i
        int sign

    elementary_symmetric_polynomial = []
    tail = float(coefficients[-1])
    size = len(coefficients)

    for i in range(size):
        sign = 1 if (i % 2) == 0 else -1
        el = sign * coefficients[size - i - 1] / tail
        elementary_symmetric_polynomial.append(el)
    return elementary_symmetric_polynomial


cdef class PolynomialParameters(object):
    cdef:
        public list elementary_symmetric_polynomial
        public list power_sum
    def __init__(self, elementary_symmetric_polynomial, power_sum):
        self.elementary_symmetric_polynomial = elementary_symmetric_polynomial
        self.power_sum = power_sum

    def __iter__(self):
        yield self.power_sum
        yield self.elementary_symmetric_polynomial


cdef class PhiConstants(object):
    cdef:
        public int order
        public Element element
        public PolynomialParameters element_coefficients
        public PolynomialParameters mass_coefficients

    def __init__(self, order, element, element_coefficients, mass_coefficients):
        self.order = order
        self.element = element
        self.element_coefficients = element_coefficients
        self.mass_coefficients = mass_coefficients


cdef class Isotope(object):
    '''
    Isotope represents an elenent with an integer number of neutrons specified.

    Attributes
    ----------
    mass: float
        Measurable mass in Da
    abundance: float [0.0:1.0]
        The abundance of this isotope in nature. This number is between 0.0 and 1.0
    neutron_shift: int
        The number of neutrons different between this isotope and the "normal" form. May be 0
        if this represents that normal form.
    '''
    cdef:
        public double mass
        public double abundance
        public int neutron_shift
        public int neutrons

    def __init__(self, mass, abundance, neutron_shift, neutrons):
        self.mass = mass
        self.abundance = abundance
        self.neutron_shift = neutron_shift
        self.neutrons = neutrons

    def __repr__(self):
        return "Isotope(mass=%0.3f, abundance=%0.3f, neutron_shift=%d, neutrons=%d)" % (
            self.mass, self.abundance, self.neutron_shift, self.neutrons)



cdef int max_variants(dict composition):
    """Calculates the maximum number of isotopic variants that could be produced by a
    composition.

    Parameters
    ----------
    composition : Mapping
        Any Mapping type where keys are element symbols and values are integers

    Returns
    -------
        max_n_variants : int
    """
    max_n_variants = 0

    for element, count in composition.items():
        if element == "H+":
            continue
        try:
            max_n_variants += count * periodic_table[element].max_neutron_shift()
        except KeyError:
            pass

    return max_n_variants


cdef double calculate_mass(dict composition, dict mass_data=None):
    cdef:
        double mass
        object match
    mass = 0.0
    if mass_data is None:
        mass_data = nist_mass
    for element in composition:
            try:
                mass += (composition[element] * mass_data[element][0][0])
            except KeyError:
                match = re.search(r"(\S+)\[(\d+)\]", element)
                if match:
                    element_ = match.group(1)
                    isotope = int(match.group(2))
                    mass += composition[element] * mass_data[element_][isotope][0]
                else:
                    raise

    return mass


cdef str _make_isotope_string(str element, int isotope=0):
    if isotope == 0:
        return element
    else:
        return "%s[%d]" % (element, isotope)


cdef tuple _get_isotope(str element_string):
    cdef:
        object match
        str element_
        int isotope
    if "[" in element_string:
        match = re.search(r"(\S+)\[(\d+)\]", element_string)
        if match:
            element_ = match.group(1)
            isotope = int(match.group(2))
            return element_, isotope
    else:
        return element_string, 0


cpdef Element make_fixed_isotope_element(Element element, int neutrons):
    cdef:
        Isotope isotope
        Element el
    isotope = element.isotopes[neutrons - element.isotopes[0].neutrons]
    el = Element(element.symbol + ("[%d]" % neutrons), dict([
            (0, Isotope(isotope.mass, abundance=1.0, neutron_shift=0, neutrons=neutrons)),
        ]))
    return el


def _isotopes_of(element):
    freqs = dict()
    for i, mass_freqs in nist_mass[element].items():
        if i == 0:
            continue
        if mass_freqs[1] > 0:
            freqs[i] = mass_freqs
    if len(freqs) == 0:
        return dict()
    mono_neutrons = max(freqs.items(), key=lambda x: x[1][1])[0]
    freqs = list(sorted(
        [(k - mono_neutrons, Isotope(*v, neutron_shift=k - mono_neutrons, neutrons=k))
                                for k, v in freqs.items()], key=lambda x: x[0]))
    return dict(freqs)


cdef class Element(object):

    def __init__(self, str symbol, dict isotopes=None):
        if isotopes is None:
            isotopes = _isotopes_of(symbol)
        self.symbol = symbol
        self.isotopes = isotopes
        min_shift = 1000
        max_shift = 0
        for shift in self.isotopes:
            if shift > max_shift:
                max_shift = shift
            if shift < min_shift:
                min_shift = shift
        self._min_neutron_shift = min_shift
        self._max_neutron_shift = max_shift
        self._no_mass_elementary_symmetric_polynomial_cache = None
        self._no_mass_power_sum_cache = None
        self._mass_elementary_symmetric_polynomial_cache = None
        self._mass_power_sum_cache = None
        try:
            self._monoisotopic_mass = self.isotopes[0].mass
        except:
            self._monoisotopic_mass = nist_mass[self.symbol][0][0]

    def __iter__(self):
        for key in sorted(self.isotopes.keys()):
            yield self.isotopes[key]

    def max_neutron_shift(self):
        return self._max_neutron_shift

    def min_neutron_shift(self):
        return self._min_neutron_shift

    cpdef double monoisotopic_mass(self):
        return self._monoisotopic_mass


periodic_table = periodic_table = {k: Element(k) for k in nist_mass}


cdef class IsotopicConstants(dict):
    cdef:
        public long _order

    def __init__(self, order):
        self._order = 0
        self.order = order

    property order:
        def __get__(self):
            return self._order

        def __set__(self, value):
            self._order = value
            self.update_coefficients()

    cdef PolynomialParameters coefficients(self, Element element, bint with_mass=False):
        cdef:
            int max_isotope_number, current_order
            list accumulator, isotope_keys
            Isotope isotope
            double coef
            size_t isotope_iter, i

        if with_mass:
            if element._mass_elementary_symmetric_polynomial_cache is not None:
                return PolynomialParameters(
                    list(element._mass_elementary_symmetric_polynomial_cache),
                    list(element._mass_power_sum_cache))
        else:
            if element._no_mass_elementary_symmetric_polynomial_cache is not None:
                return PolynomialParameters(
                    list(element._no_mass_elementary_symmetric_polynomial_cache),
                    list(element._no_mass_power_sum_cache))

        max_isotope_number = element._max_neutron_shift
        isotope_keys = sorted(element.isotopes, reverse=True)
        accumulator = []
        for isotope_iter in range(PyList_GET_SIZE(isotope_keys)):
            isotope = <Isotope>PyDict_GetItem(element.isotopes, <object>PyList_GET_ITEM(isotope_keys, isotope_iter))
            current_order = max_isotope_number - isotope.neutron_shift
            if with_mass:
                coef = isotope.mass
            else:
                coef = 1.

            if current_order > len(accumulator):
                for i in range(len(accumulator), current_order):
                    PyList_Append(accumulator, 0.)
                PyList_Append(accumulator, isotope.abundance * coef)
            elif current_order == len(accumulator):
                PyList_Append(accumulator, isotope.abundance * coef)
            else:
                raise Exception("The list of neutron shifts is not ordered.")

        elementary_symmetric_polynomial = vietes(accumulator)
        power_sum = []
        newton(power_sum, elementary_symmetric_polynomial, len(accumulator) - 1)
        if with_mass:
            if element._mass_elementary_symmetric_polynomial_cache is None:
                element._mass_elementary_symmetric_polynomial_cache = list(elementary_symmetric_polynomial)
                element._mass_power_sum_cache = list(power_sum)
        else:
            if element._no_mass_elementary_symmetric_polynomial_cache is None:
                element._no_mass_elementary_symmetric_polynomial_cache = list(elementary_symmetric_polynomial)
                element._no_mass_power_sum_cache = list(power_sum)
        return PolynomialParameters(elementary_symmetric_polynomial, power_sum)

    cdef void add_element(self, str symbol):
        cdef:
            Element element
            int order
            int isotope
            str symbol_parsed
            PolynomialParameters element_parameters, mass_parameters

        if symbol in self:
            return
        try:
            element = periodic_table[symbol]
        except KeyError:
            symbol_parsed, isotope = _get_isotope(symbol)
            if isotope == 0:
                raise KeyError(symbol)
            element = make_fixed_isotope_element(periodic_table[symbol_parsed], isotope)
        order = element.max_neutron_shift()
        element_parameters = self.coefficients(element)
        mass_parameters = self.coefficients(element, True)
        self[symbol] = PhiConstants(order, element, element_parameters, mass_parameters)

    cdef void update_coefficients(self):
        cdef:
            str symbol
            PhiConstants phi_constants
            size_t i

        for symbol, phi_constants in self.items():
            if self.order < phi_constants.order:
                continue

            for i in range(phi_constants.order, self.order + 1):
                phi_constants.element_coefficients.elementary_symmetric_polynomial.append(0.)
                phi_constants.mass_coefficients.elementary_symmetric_polynomial.append(0.)

            phi_constants.order = len(phi_constants.element_coefficients.elementary_symmetric_polynomial)
            newton(phi_constants.element_coefficients.power_sum,
                   phi_constants.element_coefficients.elementary_symmetric_polynomial,
                   phi_constants.order)
            newton(phi_constants.mass_coefficients.power_sum,
                   phi_constants.mass_coefficients.elementary_symmetric_polynomial,
                   phi_constants.order)

    cdef double nth_element_power_sum(self, str symbol, int order):
        cdef:
            PhiConstants constants
        constants = <PhiConstants>self[symbol]
        return constants.element_coefficients.power_sum[order]

    cdef double nth_modified_element_power_sum(self, str symbol, int order):
        cdef:
            PhiConstants constants
        constants = <PhiConstants>self[symbol]
        return constants.mass_coefficients.power_sum[order]


cdef class IsotopicDistribution(object):
    cdef:
        public dict composition
        public IsotopicConstants _isotopic_constants
        public int _order
        public double average_mass
        public Peak monoisotopic_peak

    def __init__(self, composition, order=-1):
        self.composition = dict(composition)
        self._isotopic_constants = IsotopicConstants(order)
        self._order = 0
        self.order = order
        self.average_mass = 0.
        self.monoisotopic_peak = self._create_monoisotopic_peak()

    property order:
        def __get__(self):
            return self._order

        def __set__(self, value):
            max_variant_count = max_variants(self.composition)
            if value == -1:
                self._order = max_variant_count
            else:
                self._order = min(value, max_variant_count)
            self._update_isotopic_constants()

    cpdef _update_isotopic_constants(self):
        cdef:
            str element
        for element in self.composition:
            self._isotopic_constants.add_element(element)
        self._isotopic_constants.order = self._order

    cdef Peak _create_monoisotopic_peak(self):
        mass = calculate_mass(self.composition)
        intensity = 0.
        for element in self.composition:
            if element == "H+":
                continue
            # intensity += log(periodic_table[element].isotopes[0].abundance)
            intensity += log(self._isotopic_constants[element].element.isotopes[0].abundance)
        intensity = exp(intensity)
        return Peak(mass, intensity, 0)

    cdef double _phi_value(self, int order):
        cdef:
            double phi
            str element
            double count
        phi = 0.
        for element, count in self.composition.items():
            if element == "H+":
                continue
            phi += self._isotopic_constants.nth_element_power_sum(element, order) * count
        return phi

    cdef double _modified_phi_value(self, str symbol, int order):
        cdef:
            double phi
            str element
            double count
            double coef
            Py_ssize_t pos
            PyObject* pk
            PyObject* pv

        phi = 0.
        pos = 0
        #for element, count in self.composition.items():
        while PyDict_Next(self.composition, &pos, &pk, &pv):
            element = <str>pk
            count = PyFloat_AsDouble(<object>pv)
            if element == "H+":
                continue
            # Count is one lower for this symbol because an isotope is present
            # accounted for in the call to `nth_modified_element_power_sum` at
            # the end?
            coef = (count if element != symbol else count - 1)
            phi += self._isotopic_constants.nth_element_power_sum(element, order) * coef

        phi += self._isotopic_constants.nth_modified_element_power_sum(symbol, order)
        return phi

    cpdef list phi_values(self):
        cdef:
            list power_sum
            size_t i
        power_sum = [0.]
        for i in range(1, self.order + 1):
            power_sum.append(self._phi_value(i))
        return power_sum

    cpdef list modified_phi_values(self, symbol):
        cdef:
            list power_sum
            size_t i
        power_sum = [0.]
        for i in range(1, self.order + 1):
            power_sum.append(self._modified_phi_value(symbol, i))
        return power_sum

    cpdef list probability(self):
        cdef:
            list phi_values
            int max_variant_count
            list probability_vector
            size_t i
            int sign

        phi_values = self.phi_values()
        max_variant_count = max_variants(self.composition)
        probability_vector = []
        newton(phi_values, probability_vector, max_variant_count)

        for i in range(0, len(probability_vector)):
            # The sign of each term in the probability vector (populated by
            # Newton's Identities by solving for the Elementary Symmetric Polynomial
            # given the Power Sums) alternates in the same order as `sign`.
            # This ensures that the probability vector is strictly positive.
            sign = 1 if i % 2 == 0 else -1
            # q(j) = q(0)  * e(j) * (-1)^j
            # intensity of the jth peak is |probability[j]| * the intensity of monoisotopic peak
            probability_vector[i] *= self.monoisotopic_peak.intensity * sign

        return probability_vector

    cpdef list center_mass(self, list probability_vector):
        cdef:
            list mass_vector
            list composition_elements
            int max_variant_count, sign
            dict ele_sym_poly_map
            str element
            list ele_sym_poly
            list power_sum
            size_t i, j
            Py_ssize_t k
            Element element_obj
            PhiConstants phi_obj
            double center, temp
            double _element_count, polynomial_term, _monoisotopic_mass
            double base_intensity
            PyObject* pk
            PyObject* pv

        mass_vector = []
        max_variant_count = max_variants(self.composition)

        base_intensity = self.monoisotopic_peak.intensity
        ele_sym_poly_map = dict()
        composition_elements = PyDict_Keys(self.composition)

        for j in range(PyList_GET_SIZE(composition_elements)):
            element = <str>PyList_GET_ITEM(composition_elements, j)
            if element == "H+":
                continue
            power_sum = self.modified_phi_values(element)
            ele_sym_poly = []
            newton(power_sum, ele_sym_poly, max_variant_count)
            ele_sym_poly_map[element] = ele_sym_poly
        for i in range(self._order + 1):
            sign = 1 if i % 2 == 0 else -1
            center = 0.0
            k = 0
            while(PyDict_Next(ele_sym_poly_map, &k, &pk, &pv)):

            #for element, ele_sym_poly in ele_sym_poly_map.items():
                element = <str>pk
                ele_sym_poly = <list>pv

                _element_count = PyFloat_AsDouble(<object>PyDict_GetItem(self.composition, element))

                polynomial_term = PyFloat_AsDouble(<object>PyList_GET_ITEM(ele_sym_poly, i))

                phi_obj = <PhiConstants>PyDict_GetItem(self._isotopic_constants, element)
                element_obj = phi_obj.element
                # element_obj = (<Element>PyDict_GetItem(periodic_table, element))
                _monoisotopic_mass = element_obj._monoisotopic_mass

                temp = _element_count
                temp *= sign * polynomial_term
                temp *= base_intensity * _monoisotopic_mass
                center += temp

            mass_vector.append((center / probability_vector[i]) if probability_vector[i] > 0 else 0)
        return mass_vector

    def aggregated_isotopic_variants(self, int charge=0, double charge_carrier=PROTON):
        '''
        Compute the m/z (or neutral mass when `charge` == 0) for each
        aggregated isotopic peak and their intensity relative to
        the monoisotopic peak.
        '''
        cdef:
            list probability_vector
            list center_mass_vector
            list peak_set
            double average_mass, adjusted_mz
            double total
            size_t i
            Peak peak
            double center_mass_i, intensity_i
        probability_vector = self.probability()
        center_mass_vector = self.center_mass(probability_vector)

        peak_set = []
        average_mass = 0.
        total = sum(probability_vector)

        for i in range(self.order + 1):
            center_mass_i = PyFloat_AsDouble(<object>PyList_GET_ITEM(center_mass_vector, i))
            if charge != 0:
                adjusted_mz = mass_charge_ratio(center_mass_i, charge, charge_carrier)
            else:
                adjusted_mz = center_mass_i

            intensity_i = PyFloat_AsDouble(<object>PyList_GET_ITEM(probability_vector, i))

            peak = Peak(adjusted_mz, intensity_i / total, charge)
            if peak.intensity < 0:
                continue
            peak_set.append(peak)
            average_mass += adjusted_mz * intensity_i

        average_mass /= total
        self.average_mass = average_mass
        peak_set.sort(key=mz_getter)
        return tuple(peak_set)



'''
A Python Implementation of the Baffling Recursive Algorithm for Isotopic cluster distributioN
'''
import os

from .brainpy import (isotopic_variants, IsotopicDistribution, periodic_table,
                      max_variants, calculate_mass, neutral_mass, mass_charge_ratio,
                      PROTON, _has_c, Peak)

from .composition import parse_formula, PyComposition


SimpleComposition = PyComposition


def get_include():
    """Retrieve the path to compiled C extensions' source files to make linking simple.

    This module contains two variants of the algorithm reimplimented using C and the Python-C API.

    The `_speedup` module is a direct translation of the pure Python implementation using Cython,
    using static typing and `cdef class` versions of the existing classes. As this implementation still
    spends a substantial amount of time in Python-space, it is slower than the option below, but is more
    straight-forward to manipulate from Python.

    The `_c` module is a complete rewrite of the algorithm directly in C, using Python only to load
    mass configuration information and is fully usable. It exports an entrypoint function to Python
    which replaces the :func:`isotopic_variants` function when available. Because almost all of the
    action happens in C here, it's not possible to run individual parts of the process directly from
    Python.
    """
    return os.path.join(__path__[0], "_c")


if _has_c:
    from .brainpy import _IsotopicDistribution

__author__ = "Joshua Klein & Han Hu"


__all__ = [
    "isotopic_variants", "IsotopicDistribution", "periodic_table",
    "max_variants", "calculate_mass", "neutral_mass", "mass_charge_ratio",
    "PROTON", "_has_c", "Peak",

    "parse_formula", "PyComposition", "SimpleComposition",

    "_IsotopicDistribution",

    "get_include"
]



# cython: embedsignature=True

cimport cython

# from cpython.string cimport PyString_FromString, PyString_AsString
# from cpython.int cimport PyInt_AsLong, PyInt_FromLong

from brainpy._c.compat cimport (
    PyStr_FromString as PyString_FromString, PyStr_AsString as PyString_AsString,
    PyInt_AsLong, PyInt_FromLong)

from cpython.list cimport PyList_New, PyList_Append, PyList_Append
from cpython.dict cimport PyDict_SetItem, PyDict_GetItem
from cpython.object cimport PyObject
from libc.string cimport strcmp, memcpy, strlen, strncpy
from libc.stdlib cimport malloc, free, realloc, atoi, calloc
from libc.math cimport abs, fabs
from libc cimport *

cdef extern from * nogil:
    int printf (const char *template, ...)
    int sprintf(char *, char *,...)


from brainpy.mass_dict import nist_mass as __nist_mass

cdef dict nist_mass
nist_mass = __nist_mass


cdef double PROTON = nist_mass["H+"][0][0]


cdef double neutral_mass(double mz,  int z, double charge_carrier=PROTON) nogil:
    return (mz * fabs(z)) - (z * charge_carrier)


@cython.cdivision
cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=PROTON) nogil:
    return (neutral_mass + (z * charge_carrier)) / fabs(z)


cdef char* _parse_isotope_string(char* label, int* isotope_num, char* element_name) nogil:
    cdef:
        size_t i = 0
        bint in_bracket = False
        char current
        size_t name_end
        size_t num_start
        size_t num_end

        size_t size
        char[4] number_part

    name_end = 0
    num_start = 0
    num_end = 0
    size = strlen(label)
    for i in range(size):
        current = label[i]
        if in_bracket:
            if current == 93:
                break
            num_end += 1
        elif current == 91:
            in_bracket = True
            name_end = i
            num_start = i + 1
            num_end = num_start
        else:
            name_end += 1

    if num_start > 0:
        memcpy(number_part, label + num_start, num_end - num_start)
        number_part[num_end - num_start] = '\0'
        isotope_num[0] = atoi(number_part)
    else:
        isotope_num[0] = 0

    element_name[name_end] = '\0'
    memcpy(element_name, label, name_end)

    return element_name

cdef char* _make_isotope_string(Element* element, Isotope* isotope, char* out) nogil:
    if isotope.neutron_shift == 0:
        sprintf(out, "%s", element.symbol)
        return out
    else:
        sprintf(out, "%s[%d]", element.symbol, isotope.neutrons)
        return out

cdef char* _make_fixed_isotope_string(Element* element, Isotope* isotope, char* out) nogil:
    sprintf(out, "%s[%d]", element.symbol, isotope.neutrons)
    return out

# -----------------------------------------------------------------------------
# Isotope and IsotopeMap Methods

cdef IsotopeMap* make_isotope_map(list organized_isotope_data, size_t size):
    cdef:
        IsotopeMap* result
        size_t i
        int k
        tuple values

    result = <IsotopeMap*>malloc(sizeof(IsotopeMap))
    result.bins = <Isotope*>malloc(sizeof(Isotope) * size)
    result.size = size

    i = 0
    for k, values in organized_isotope_data:
        result.bins[i].mass = values[0]
        result.bins[i].abundance = values[1]
        result.bins[i].neutron_shift = values[2]
        result.bins[i].neutrons = values[3]
        i += 1
    return result


cdef Isotope* get_isotope_by_neutron_shift(IsotopeMap* isotopes, int neutron_shift) nogil:
    cdef:
        size_t i
        Isotope* isotope_item

    for i in range(isotopes.size):
        isotope_item = &(isotopes.bins[i])
        # if isotope_item.neutron_shift == neutron_shift:
        if isotopes.bins[i].neutron_shift == neutron_shift:
            return isotope_item
    return NULL


cdef int get_isotope_by_neutron_shift_index(IsotopeMap* isotopes, int neutron_shift) nogil:
    cdef:
        size_t i

    for i in range(isotopes.size):
        if isotopes.bins[i].neutron_shift == neutron_shift:
            return i
    return -1


cdef Isotope* get_isotope_by_neutron_count(IsotopeMap* isotopes, int neutrons) nogil:
    cdef:
        size_t i
        Isotope* isotope_item

    for i in range(isotopes.size):
        isotope_item = &(isotopes.bins[i])
        if isotope_item.neutrons == neutrons:
            return isotope_item
    return NULL


cdef void print_isotope_map(IsotopeMap* isotope_map):
    cdef:
        size_t i

    for i in range(isotope_map.size):
        printf("%f, %f, %d\n", isotope_map.bins[i].mass, isotope_map.bins[i].abundance, isotope_map.bins[i].neutron_shift)


cdef void free_isotope_map(IsotopeMap* isotopes) nogil:
    free(isotopes.bins)
    free(isotopes)


# -----------------------------------------------------------------------------
# Element Methods

cdef double element_monoisotopic_mass(Element* element) nogil:
    if element.monoisotopic_isotope_index < 0:
        return get_isotope_by_neutron_shift(element.isotopes, 0).mass
    else:
        return element.isotopes.bins[element.monoisotopic_isotope_index].mass

cdef double element_isotopic_mass(Element* element, int isotope_number) nogil:
    return get_isotope_by_neutron_count(element.isotopes, isotope_number).mass

cdef int element_min_neutron_shift(Element* element) nogil:
    return element.isotopes.bins[0].neutron_shift

cdef int element_max_neutron_shift(Element* element) nogil:
    return element.isotopes.bins[element.isotopes.size - 1].neutron_shift

def __select_1_1(object x):
    return x[1][1]

cdef void _isotopes_of(char* element_symbol, IsotopeMap** isotope_frequencies):
    cdef:
        dict freqs, element_data
        list freq_list
        int i, key
        tuple mass, value
        double abundance
        list bunches
        tuple mass_freqs
        int mono_neutrons
        object k, v
        str py_element_symbol

    freqs = dict()
    py_element_symbol = PyString_FromString(element_symbol)

    try:
        element_data = nist_mass[py_element_symbol]
    except KeyError:
        element_data = nist_mass[py_element_symbol.encode('utf-8')]

    for i, mass_freqs in element_data.items():
        if i == 0:
            continue
        if mass_freqs[1] > 0:
            freqs[i] = mass_freqs

    if len(freqs) == 0:
        isotope_frequencies[0] = make_isotope_map([], 0)
        return

    freq_list = list(freqs.items())
    freq_list.sort(key=__select_1_1)
    mono_neutrons = freq_list[-1][0]

    bunches = []
    for k, v in freqs.items():
        bunches.append((k - mono_neutrons, (v[0], v[1], k - mono_neutrons, k)))

    bunches.sort()

    # bunches = sorted([((k - mono_neutrons, (v[0], v[1], k - mono_neutrons, k))
    #                   for k, v in freqs.items())], key=lambda x: x[0])
    isotope_frequencies[0] = make_isotope_map(bunches, len(bunches))

cdef Element* make_element(char* symbol):
    cdef:
        Element* element
    element = <Element*>malloc(sizeof(Element))
    element.symbol = symbol
    _isotopes_of(symbol, &element.isotopes)
    element.monoisotopic_isotope_index = -1
    element.monoisotopic_isotope_index = get_isotope_by_neutron_shift_index(
        element.isotopes, 0)
    return element

cdef void free_element(Element* element) nogil:
    free_isotope_map(element.isotopes)
    free(element)

cdef Element* make_fixed_isotope_element(Element* element, int neutron_count) nogil:
    cdef:
        int i
        size_t j
        Element* out
        IsotopeMap* isotope_map
        Isotope* reference
        char* symbol
    out = <Element*>malloc(sizeof(Element))
    isotope_map = <IsotopeMap*>malloc(sizeof(IsotopeMap))
    isotope_map.bins = <Isotope*>malloc(sizeof(Isotope) * 1)
    isotope_map.size = 1
    out.isotopes = isotope_map
    reference = get_isotope_by_neutron_count(element.isotopes, neutron_count)
    if reference == NULL:
        printf("reference was NULL!\n")
        return NULL
    isotope_map.bins[0].mass = reference.mass
    isotope_map.bins[0].abundance = 1.0
    isotope_map.bins[0].neutron_shift = 0
    isotope_map.bins[0].neutrons = neutron_count
    out.symbol = <char*>malloc(sizeof(char) * 10)
    _make_fixed_isotope_string(element, reference, out.symbol)
    out.monoisotopic_isotope_index = 0
    return out


# -----------------------------------------------------------------------------
# ElementHashTable and ElementHashBucket Methods

cdef size_t hash_string(char *string_) nogil:
    cdef:
        size_t hash_value
        size_t i
        int c
    hash_value = 5381
    i = 0
    c = string_[i]
    while (c):
        hash_value = ((hash_value << 5) + hash_value) + c
        i += 1
        c = string_[i]
    return hash_value;


def test_hash_string(str string_):
    return hash_string(PyString_AsString(string_))


cdef ElementHashTable* make_element_hash_table(size_t size) nogil:
    cdef:
        ElementHashTable* table
        size_t i

    table = <ElementHashTable*>malloc(sizeof(ElementHashTable))
    table.buckets = <ElementHashBucket*>malloc(sizeof(ElementHashBucket) * size)
    table.size = size

    for i in range(size):
        table.buckets[i].size = 6
        table.buckets[i].elements = <Element**>malloc(sizeof(Element*) * table.buckets[i].size)
        table.buckets[i].used = 0
    return table


cdef int element_hash_bucket_resize(ElementHashBucket* bucket) nogil:
    cdef:
        Element** elements
        size_t new_size

    new_size = bucket.size * 2
    elements = <Element**>realloc(bucket.elements, sizeof(Element*) * new_size)
    if elements == NULL:
        printf("element_hash_bucket_resize failed\n")
        return -1
    bucket.elements = elements
    return 0


cdef int element_hash_bucket_insert(ElementHashBucket* bucket, Element* element) nogil:
    cdef:
        size_t i
        int status
        Element* el
    if (bucket.used + 1) == bucket.size:
        status = element_hash_bucket_resize(bucket)
        if status != 0:
            printf("element_hash_bucket_insert failed with %s\n", element.symbol)
            return -1
    # printf("Inserting %s into bucket with total capacity %d\n", element.symbol, bucket.size)
    bucket.elements[bucket.used] = element
    # printf("Element %s put in slot %d\n", bucket.elements[bucket.used].symbol, bucket.used)
    bucket.used += 1
    return 0


cdef int element_hash_bucket_find(ElementHashBucket* bucket, char* symbol, Element** out) nogil:
    cdef:
        size_t i
    for i in range(bucket.used):
        if strcmp(bucket.elements[i].symbol, symbol) == 0:
            out[0] = bucket.elements[i]
            return 0
    return -1


cdef void free_element_hash_bucket(ElementHashBucket* bucket) nogil:
    free(bucket.elements)
    free(bucket)


@cython.cdivision(True)
cdef int element_hash_table_get(ElementHashTable* table, char* symbol, Element** out) nogil:
    cdef:
        size_t hash_value
        size_t position
        ElementHashBucket bucket
        int status

    hash_value = hash_string(symbol)
    position = hash_value % table.size
    bucket = (table.buckets[position])
    status = element_hash_bucket_find(&bucket, symbol, out)
    return status


@cython.cdivision(True)
cdef int element_hash_table_put(ElementHashTable* table, Element* element) nogil:
    cdef:
        size_t hash_value
        size_t position
        ElementHashBucket bucket
        int status

    hash_value = hash_string(element.symbol)
    position = hash_value % table.size
    bucket = (table.buckets[position])
    status = element_hash_bucket_insert(&bucket, element)
    table.buckets[position] = bucket
    return status


cdef size_t free_element_hash_table(ElementHashTable* table) nogil:
    cdef:
        size_t i

    for i in range(table.size):
        free_element_hash_bucket(&(table.buckets[i]))
    free(table)

cdef ElementHashTable* make_element_hash_table_populated(size_t size):
    cdef:
        Element* element
        Element* out_test
        ElementHashTable* table
        str pk  #!
        char* k
        size_t i
        int status

    table = make_element_hash_table(size)

    for pk in nist_mass:
        k = PyString_AsString(pk)
        element = make_element(k)
        status = element_hash_table_put(table, element)
        if status != 0:
            printf("element_hash_table_put exited with status %d\n", status)
        status = element_hash_table_get(table, k, &out_test)
        if status != 0:
            printf("element_hash_table_get exited with status %d\n", status)
    return table


_ElementTable = make_element_hash_table_populated(256)


cdef ElementHashTable* get_system_element_hash_table() nogil:
    return _ElementTable

cdef int set_system_element_hash_table(ElementHashTable* table) nogil:
    _ElementTable[0] = table[0]


# -----------------------------------------------------------------------------
# Composition Methods

cdef Composition* make_composition() nogil:
    '''
    Create a new, empty Composition struct
    '''
    cdef:
        Composition* composition
    composition = <Composition*>malloc(sizeof(Composition))
    composition.elements = <char**>calloc(7, sizeof(char*))
    composition.counts = <count_type*>calloc(7, sizeof(count_type))
    composition.size = 7
    composition.used = 0
    composition.max_variants = 0
    return composition

cdef int composition_eq(Composition* composition_1, Composition* composition_2) nogil:
    '''
    Test two Composition instances for element-wise equality
    '''
    cdef:
        int status
        size_t i
        char* symbol
        count_type value_1, value_2

    if composition_1.used != composition_2.used:
        return 0
    i = 0
    while i < composition_2.used:
        symbol = composition_2.elements[i]
        status = composition_get_element_count(composition_2, symbol, &value_2)
        if status != 0:
            return 0
        status = composition_get_element_count(composition_1, symbol, &value_1)
        if status != 0:
            return 0
        if value_1 != value_2:
            return 0
        i += 1

    return 1

cdef Composition* copy_composition(Composition* composition) nogil:
    '''
    Create a new Composition instance whose element counts are copied from
    `composition`
    '''
    cdef:
        Composition* result
        int status
        size_t i
    result = <Composition*>malloc(sizeof(Composition))
    result.elements = <char**>malloc(sizeof(char*) * composition.size)
    result.counts = <count_type*>malloc(sizeof(count_type) * composition.size)
    result.size = composition.size
    result.used = 0

    i = 0
    while i < composition.used:
        status = composition_set_element_count(result, composition.elements[i], composition.counts[i])
        i += 1
    return result

cdef void print_composition(Composition* composition) nogil:
    cdef:
        size_t i
        int status
    i = 0
    printf("Addr %d %d\n", <int>composition, <int>(composition.elements))
    printf("Size: %d, Used: %d\n", composition.size, composition.used)
    printf("{\n")
    while i < composition.used:
        printf("  %s: %d\n", composition.elements[i], composition.counts[i])
        i += 1
    printf("}\n\n")

cdef int composition_set_element_count(Composition* composition, char* element, count_type count) nogil:
    '''
    Set the count for `element` in `composition`

    Return Values:
    0: Success
    1: General Failure
    -1: Failure due to Out-of-Memory
    '''
    cdef:
        size_t i
        int status
        bint done
    done = False
    i = 0
    while (i < composition.used):
        if strcmp(element, composition.elements[i]) == 0:
            done = True
            composition.counts[i] = count
            return 0
        i += 1

    if not done:
        composition.used += 1
        if composition.used >= composition.size:
            status = composition_resize(composition)
            if status != 0:
                return -1
        composition.elements[i] = element
        composition.counts[i] = count
        return 0
    return 1

cdef int composition_get_element_count(Composition* composition, char* element, count_type* count) nogil:
    '''
    Get the count of `element` in `composition`. The count is 0 if `element` is not in `composition`

    Return Values:
    0: Success
    1: Not Found
    '''
    cdef:
        size_t i
        int status
        bint done
    done = False
    i = 0
    while (i < composition.used):
        if strcmp(element, composition.elements[i]) == 0:
            done = True
            count[0] = composition.counts[i]
            return 0
        i += 1

    if not done:
        count[0] = 0
    return 1

cdef int composition_inc_element_count(Composition* composition, char* element, count_type increment) nogil:
    '''
    Increase the count for `element` in `composition` by `increment`.

    Return Values:
    0: Success
    1: General Failure
    -1: Failure due to Out-of-Memory
    '''
    cdef:
        size_t i
        int status
        bint done
    done = False
    i = 0
    while (i < composition.used):
        if strcmp(element, composition.elements[i]) == 0:
            done = True
            composition.counts[i] += increment
            return 0
        i += 1

    if not done:
        composition.used += 1
        if composition.used >= composition.size:
            status = composition_resize(composition)
            if status != 0:
                return -1
        composition.elements[i] = element
        composition.counts[i] = increment
        return 0
    return 1

cdef int composition_resize(Composition* composition) nogil:
    '''
    Increases the size of the parallel arrays in `composition`, doubling them in length

    Return Values:
    0: Success
    -1: Failure due to Out-of-Memory
    '''
    composition.elements = <char**>realloc(composition.elements, sizeof(char*) * composition.size * 2)
    composition.counts = <count_type*>realloc(composition.counts, sizeof(count_type) * composition.size * 2)
    composition.size *= 2
    if composition.counts == NULL:
        return -1
    return 0

cdef double composition_mass(Composition* composition) nogil:
    '''
    Calculates the monoisotopic mass of `composition`
    '''
    cdef:
        double mass
        Element* element
        char* element_label
        char[10] element_name
        int isotope_number
        int status
        size_t i
    i = 0
    while i < composition.used:
        element_label = composition.elements[i]
        _parse_isotope_string(element_label, &isotope_number, element_name)
        status = element_hash_table_get(_ElementTable, element_name, &element)
        if status != 0:
            printf("Could not find element %s\n", element_label)
        if isotope_number == 0:
            mass += element_monoisotopic_mass(element) * composition.counts[i]
        else:
            mass += element_isotopic_mass(element, isotope_number) * composition.counts[i]
        i += 1

    return mass

cdef void free_composition(Composition* composition) nogil:
    free(composition.elements)
    free(composition.counts)
    free(composition)

cdef Composition* composition_add(Composition* composition_1, Composition* composition_2, int sign) nogil:
    cdef:
        Composition* result
        count_type value
        char* symbol
        int status
        size_t i

    i = 0
    result = copy_composition(composition_1)

    while i < composition_2.used:
        symbol = composition_2.elements[i]
        status = composition_get_element_count(composition_2, symbol, &value)
        i += 1
        if value == 0:
            continue
        status = composition_inc_element_count(result, symbol, sign * value)
    return result

cdef int composition_iadd(Composition* composition_1, Composition* composition_2, int sign) nogil:
    cdef:
        char* symbol
        count_type value
        int status
        size_t i
    i = 0
    status = 0
    while i < composition_2.used:
        symbol = composition_2.elements[i]
        status = composition_get_element_count(composition_2, symbol, &value)
        i += 1
        if value == 0:
            continue
        status = composition_inc_element_count(composition_1, symbol, sign * value)
        if status != 0:
            pass
    return status

cdef Composition* composition_mul(Composition* composition, int scale) nogil:
    cdef:
        Composition* result
        size_t i

    i = 0
    result = copy_composition(composition)
    while i < result.used:
        result.counts[i] *= scale
        i += 1
    return result

cdef void composition_imul(Composition* composition, int scale) nogil:
    cdef:
        size_t i

    i = 0
    while i < composition.used:
        composition.counts[i] *= scale
        i += 1

cdef dict composition_to_dict(Composition* composition):
    cdef:
        dict result
        str symbol
        size_t i
        char* symbol_c
        count_type value
    i = 0
    result = dict()
    while i < composition.used:
        symbol_c = composition.elements[i]
        composition_get_element_count(composition, symbol_c, &value)
        if value == 0:
            continue
        symbol = PyString_FromString(symbol_c)
        result[symbol] = value
        i += 1
    return result

cdef Composition* dict_to_composition(dict comp_dict):
    cdef:
        Composition* result
        str symbol
        size_t i
        char* symbol_c
        count_type value
    result = make_composition()
    for symbol, value in comp_dict.items():
        symbol_c = PyString_AsString(symbol)
        composition_set_element_count(result, symbol_c, value)
    return result


# -----------------------------------------------------------------------------
# Prototyping and Testing

def main():
    cdef Element* elem
    cdef Composition* composition
    cdef Composition* sum_composition
    cdef int i, j
    cdef char *sym
    cdef char *k
    cdef char[10] element_name_buffer
    cdef dict copy

    sym = "O"

    j = element_hash_table_get(_ElementTable, "C", &elem)
    # print elem.symbol, element_monoisotopic_mass(elem)

    composition = make_composition()
    composition_set_element_count(composition, "O[18]", 1)
    composition_set_element_count(composition, "H", 2)
    print_composition(composition)
    print("Massing")
    print(composition_mass(composition))

    sum_composition = composition_add(composition, composition, 1)
    print("Summed")
    print(composition_mass(sum_composition))
    print(composition_mass(composition))
    composition_iadd(sum_composition, composition, 1)
    print(composition_mass(sum_composition))


def isotope_update_test(str element_symbol):
    cdef:
        char* c_element_symbol
        char* element
        char[10] isotope_string
        int isotope
        int status
        Element* elem_obj
        Element* fixed_isotope_element
        Isotope* isotope_obj
        str out

    element = <char*>malloc(sizeof(char) * 10)
    c_element_symbol = PyString_AsString(element_symbol)
    printf("c_element_symbol %s\n", c_element_symbol)
    _parse_isotope_string(c_element_symbol, &isotope, element)
    printf("status %s\n", element)
    out = PyString_FromString(element)
    printf("Element Symbol: %s, Isotope: %d\n", element, isotope)
    print("element_hash_table_get")
    status = element_hash_table_get(_ElementTable, element, &elem_obj)
    if status != 0:
        print("Error, ", element_symbol, element, isotope)
    print("Fetched element", elem_obj.symbol)
    isotope_obj = get_isotope_by_neutron_count(elem_obj.isotopes, isotope)
    print(isotope_obj.mass, isotope_obj.neutrons, isotope_obj.neutron_shift)
    _make_isotope_string(elem_obj, isotope_obj, isotope_string)
    print(isotope_string)
    fixed_isotope_element = make_fixed_isotope_element(elem_obj, isotope)
    printf("Fixed: %s, %f", fixed_isotope_element.symbol, element_monoisotopic_mass(fixed_isotope_element))


def isotope_parse_test(str element_symbol):
    cdef:
        char* c_element_symbol
        char* element
        char[10] isotope_string
        int isotope
    element = <char*>malloc(sizeof(char) * 10)
    c_element_symbol = PyString_AsString(element_symbol)
    _parse_isotope_string(c_element_symbol, &isotope, element)
    print(element, isotope)
    free(element)


cdef dict _string_table = {}

cdef str _store_string(str symbol):
    cdef:
        PyObject* po
    po = PyDict_GetItem(_string_table, symbol)
    if po == NULL:
        PyDict_SetItem(_string_table, symbol, symbol)
        po = PyDict_GetItem(_string_table, symbol)
    return <str>po


cdef class PyComposition(object):
    '''A mapping representing a chemical composition.

    Implements arithmetic operations, +/- is defined
    between a :class:`PyComposition` and a :class:`Mapping`-like
    object, and * is defined between a :class:`PyComposition` and
    an integer.
    '''

    @staticmethod
    cdef PyComposition _create(Composition* base):
        cdef:
            PyComposition inst

        inst = PyComposition.__new__(PyComposition)
        if base != NULL:
            composition_iadd(inst.impl, base, 1)
        inst._clean = False
        return inst

    cdef void _set_impl(self, Composition* composition, bint free_existing=1):
        if free_existing:
            free_composition(self.impl)
        self.impl = composition
        self._clean = False

    def __init__(self, base=None, **kwargs):
        self.impl = make_composition()
        self._clean = False
        self.cached_mass = 0.

        if base is not None and isinstance(base, dict):
            self.update(base)

        for py_element, count in kwargs.items():
            self[py_element] = count

    def __dealloc__(self):
        free_composition(self.impl)

    def __getitem__(self, str key):
        cdef:
            count_type count
            int status
            char* ckey
        key = _store_string(key)
        ckey = PyString_AsString(key)
        status = composition_get_element_count(self.impl, ckey, &count)
        if status == 0:
            return count
        else:
            # composition_set_element_count(self.impl, ckey, 0)
            return 0.

    def __setitem__(self, str key, count_type count):
        cdef:
            int status
            char* ckey
        key = _store_string(key)
        ckey = PyString_AsString(key)
        status = composition_set_element_count(self.impl, ckey, count)
        self._clean = False

    def update(self, arg=None, **kwargs):
        if arg is not None:
            self.update(**arg)
        for key, value in kwargs.items():
            self[key] = value

    def keys(self):
        cdef:
            size_t i
            int count
            char* elem
            list keys
            str key_str
        i = 0
        keys = []
        while i < self.impl.used:
            elem = self.impl.elements[i]
            count = self.impl.counts[i]
            i += 1
            if count == 0:
                continue
            key_str = PyString_FromString(elem)
            PyList_Append(keys, key_str)
        return keys

    def values(self):
        cdef:
            size_t i
            count_type value
            list counts
        counts = []
        i = 0
        while i < self.impl.used:
            value = self.impl.counts[i]
            i += 1
            if value == 0:
                continue
            counts.append(value)
        return counts

    def pop(self, str key, object default=None):
        value = self[key]
        self[key] = 0
        if value == 0:
            return default
        else:
            return value

    def __contains__(self, str key):
        return self[key] != 0

    def items(self):
        cdef:
            size_t i
            char* elem
            count_type value
            list items

        i = 0
        items = []
        while i < self.impl.used:
            elem = self.impl.elements[i]
            value = self.impl.counts[i]
            i += 1
            if value == 0:
                continue
            items.append((PyString_FromString(elem), value))
        return items

    cpdef PyComposition copy(self):
        inst = PyComposition()
        composition_iadd(inst.impl, self.impl, 1)
        return inst

    def __add__(self, other):
        cdef:
            PyComposition result
            PyComposition _other

        result = self.copy()

        if isinstance(other, dict):
            for key, value in other.items():
                result[key] += value
        elif isinstance(other, PyComposition):
            _other = other
            composition_iadd(result.impl, _other.impl, 1)
        else:
            return NotImplemented
        return result

    def __iadd__(self, other):
        cdef:
            PyComposition _other

        if isinstance(other, dict):
            for key, value in other.items():
                self[key] += value
        elif isinstance(other, PyComposition):
            _other = other
            composition_iadd(self.impl, _other.impl, 1)
        else:
            raise TypeError("Cannot add %s to PyComposition" % type(other))
        return self

    def __sub__(self, other):
        cdef:
            PyComposition result
            PyComposition _other

        result = self.copy()

        if isinstance(other, dict):
            for key, value in other.items():
                result[key] -= value
        elif isinstance(other, PyComposition):
            _other = other
            composition_iadd(result.impl, _other.impl, -1)
        else:
            return NotImplemented
        return result

    def __isub__(self, other):
        cdef:
            PyComposition _other

        if isinstance(other, dict):
            for key, value in other.items():
                self[key] -= value
        elif isinstance(other, PyComposition):
            _other = other
            composition_iadd(self.impl, _other.impl, -1)
        else:
            raise TypeError("Cannot subtract %s from PyComposition" % type(other))
        return self

    def __mul__(self, scale):
        cdef:
            PyComposition result
            PyComposition inst
            int scale_factor

        if isinstance(self, PyComposition):
            inst = self
            scale_factor = PyInt_AsLong(scale)
        else:
            inst = scale
            scale_factor = PyInt_AsLong(self)

        result = inst.copy()

        composition_imul(result.impl, scale_factor)

        return result

    def __imul__(self, int scale):
        composition_imul(self.impl, scale)
        return self

    def __iter__(self):
        return iter(self.keys())

    cpdef double mass(self):
        '''Calculate the monoisotopic mass of this chemical composition

        Returns
        -------
        float
        '''
        if self._clean:
            return self.cached_mass
        else:
            self.cached_mass = composition_mass(self.impl)
            self._clean = True
            return self.cached_mass

    def __len__(self):
        return self.impl.used

    def __repr__(self):
        return "PyComposition({%s})" % ', '.join("\"%s\": %d" % kv for kv in self.items())

    cpdef bint __equality_pycomposition(self, PyComposition other):
        return composition_eq(self.impl, other.impl)

    cpdef bint __equality_dict(self, dict other):
        cdef:
            size_t n
            str key
            count_type value
            count_type my_value

        n = len(other)
        if n != self.impl.used:
            return False
        for key, value in other.items():
            my_value = self[key]
            if my_value != value:
                return False
        return True

    def __richcmp__(self, object other, int code):
        if not isinstance(self, PyComposition):
            other, self = self, other
        if isinstance(other, PyComposition):
            if code == 2:
                return self.__equality_pycomposition(other)
            elif code == 3:
                return not self.__equality_pycomposition(other)
        elif isinstance(other, dict):
            if code == 2:
                return self.__equality_dict(other)
            elif code == 3:
                return not self.__equality_dict(other)
        else:
            other = dict(other)
            if code == 2:
                return self.__equality_dict(other)
            elif code == 3:
                return not self.__equality_dict(other)


cdef extern from "<stdlib.h>" nogil:
    int isdigit( int ch );


cdef enum:
      START
      ISOTOPE
      ELEMENT
      COUNT


cpdef PyComposition parse_formula(str formula):
    """Parse a chemical formula and construct a :class:`PyComposition` object

    Parameters
    ----------
    formula : :class:`str`

    Returns
    -------
    :class:`PyComposition`

    Raises
    ------
    ValueError
        If the formula doesn't match the expected pattern
    """
    cdef:
        ssize_t i, n
        ssize_t elstart, elend
        ssize_t numstart, numend
        ssize_t isostart, isoend
        char* cstr
        char* temp
        char a
        Element* elem
        Element* fixed_isotope_elem
        PyComposition composition
        int state, count, prev_count, fixed_isotope, status, found
    prev_count = 0
    state = ELEMENT
    cstr = PyString_AsString(formula)
    temp = <char*>malloc(sizeof(char) * 10)
    for i in range(10):
        temp[i] = "\0"
    n = len(formula)
    elstart = isostart = isoend = 0
    composition = PyComposition()
    for i in range(n):
        a = cstr[i]
        if a == ']':
            if state == ISOTOPE:
                elend = i
                isoend = i
                numstart = i + 1
                state = COUNT
            else:
                raise ValueError((formula, i, state))
        elif a == '[':
            if state == ELEMENT:
                elend = i
                isostart = i
                state = ISOTOPE
            else:
                raise ValueError((formula, i, state))
        elif isdigit(a):
            if state == ISOTOPE or state == COUNT:
                pass
            elif state == ELEMENT:
                elend = i
                numstart = i
                state = COUNT
            else:
                raise ValueError((formula, i, state))
        else:
            if state == ELEMENT:
                continue
            elif state == COUNT:
                numend = i
                strncpy(temp, cstr + numstart, numend - numstart)
                temp[numend - numstart] = 0
                count = atoi(temp)
                strncpy(temp, cstr + elstart, elend - elstart)
                temp[elend - elstart] = 0
                found = element_hash_table_get(_ElementTable, temp, &elem)
                # something in this loop is suspicious and causes periodic crashes.
                if isostart != 0 and isoend != 0 and found != 0:
                    isostart += 1
                    strncpy(temp, cstr + isostart, isoend - isostart)
                    temp[isoend - isostart] = 0
                    fixed_isotope = atoi(temp)
                    
                    strncpy(temp, cstr + elstart, isoend - elstart + 1)
                    temp[isoend - elstart + 2] = 0

                    found = element_hash_table_get(_ElementTable, temp, &fixed_isotope_elem)
                    if found != 0:
                        fixed_isotope_elem = make_fixed_isotope_element(elem, fixed_isotope)
                        element_hash_table_put(_ElementTable, fixed_isotope_elem)
                        elem = fixed_isotope_elem
                    else:
                        elem = fixed_isotope_elem
                prev_count = 0
                composition_get_element_count(composition.impl, elem.symbol, &prev_count)
                composition_set_element_count(composition.impl, elem.symbol, count + prev_count)
                state = ELEMENT
                elstart = i
                isostart = 0
                isoend = 0

    if state == COUNT:
        numend = i + 1
        strncpy(temp, cstr + numstart, numend - numstart)
        temp[numend - numstart] = 0
        count = atoi(temp)
        strncpy(temp, cstr + elstart, elend - elstart)
        temp[elend - elstart] = 0
        found = element_hash_table_get(_ElementTable, temp, &elem)
        if isostart != 0 and isoend != 0 and found != 0:
            isostart += 1
            strncpy(temp, cstr + isostart, isoend - isostart)
            temp[isoend - isostart] = 0
            
            fixed_isotope = atoi(temp)
            
            strncpy(temp, cstr + elstart, isoend - elstart + 1)
            temp[isoend - elstart + 2] = 0

            found = element_hash_table_get(_ElementTable, temp, &fixed_isotope_elem)
            if found != 0:
                fixed_isotope_elem = make_fixed_isotope_element(elem, fixed_isotope)
                element_hash_table_put(_ElementTable, fixed_isotope_elem)
                elem = fixed_isotope_elem
            else:
                elem = fixed_isotope_elem
        prev_count = 0
        composition_get_element_count(composition.impl, elem.symbol, &prev_count)
        composition_set_element_count(composition.impl, elem.symbol, count + prev_count)

    free(temp)
    return composition



cdef dict nist_mass
cdef double PROTON
cdef dict periodic_table


cdef double neutral_mass(double mz,  int z, double charge_carrier=*)
cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*)

cdef double calculate_mass(dict composition, dict mass_data=*)

cpdef Element make_fixed_isotope_element(Element element, int isotope_shift)

cdef class Element:
    cdef:
        public str symbol
        public dict isotopes
        double _monoisotopic_mass
        int _max_neutron_shift
        int _min_neutron_shift
        list _no_mass_elementary_symmetric_polynomial_cache
        list _no_mass_power_sum_cache
        list _mass_elementary_symmetric_polynomial_cache
        list _mass_power_sum_cache
    
    cpdef double monoisotopic_mass(self)


import re

from collections import Counter

from .mass_dict import nist_mass

_isotope_string = r'^([A-Z][a-z+]*)(?:\[(\d+)\])?$'
_atom = r'([A-Z][a-z+]*)(?:\[(\d+)\])?([+-]?\d+)?'
_formula = r'^({})*$'.format(_atom)
atom_pattern = re.compile(_atom)
formula_pattern = re.compile(_formula)


def calculate_mass(composition, mass_data=None):
    """Calculates the monoisotopic mass of a composition

    Parameters
    ----------
    composition : Mapping
        Any Mapping type where keys are element symbols and values are integers
    mass_data : dict, optional
        A dict with the masses of the chemical elements (the default
        value is :py:data:`nist_mass`).

    Returns
    -------
        mass : float
    """
    mass = 0.0
    if mass_data is None:
        mass_data = nist_mass
    for element in composition:
        try:
            mass += (composition[element] * mass_data[element][0][0])
        except KeyError:
            match = re.search(r"(\S+)\[(\d+)\]", element)
            if match:
                element_ = match.group(1)
                isotope = int(match.group(2))
                mass += composition[element] * mass_data[element_][isotope][0]
            else:
                raise
    return mass


def _make_isotope_string(element, isotope=0):
    if isotope == 0:
        return element
    else:
        return "%s[%d]" % (element, isotope)


def _get_isotope(element_string):
    if "[" in element_string:
        match = re.search(r"(\S+)\[(\d+)\]", element_string)
        if match:
            element_ = match.group(1)
            isotope = int(match.group(2))
            return element_, isotope
    else:
        return element_string, 0


class PyComposition(Counter):
    '''A mapping representing a chemical composition.

    Implements arithmetic operations, +/- is defined
    between a :class:`PyComposition` and a :class:`Mapping`-like
    object, and * is defined between a :class:`PyComposition` and
    an integer.
    '''
    def __init__(self, base=None, **kwargs):
        if base is not None:
            self.update(base)
        else:
            if kwargs:
                self.update(kwargs)

    def __missing__(self, key):
        return 0

    def __mul__(self, i):
        inst = self.copy()
        for key, value in self.items():
            inst[key] = value * i
        return inst

    def __imul__(self, i):
        for key, value in list(self.items()):
            self[key] = value * i
        return self

    def mass(self, mass_data=None):
        '''Calculate the monoisotopic mass of this chemical composition

        Returns
        -------
        float
        '''
        return calculate_mass(self, mass_data)


def parse_formula(formula):
    """Parse a chemical formula and construct a :class:`PyComposition` object

    Parameters
    ----------
    formula : :class:`str`

    Returns
    -------
    :class:`PyComposition`

    Raises
    ------
    ValueError
        If the formula doesn't match the expected pattern
    """
    if not formula_pattern.match(formula):
        raise ValueError("%r does not look like a formula" % (formula,))
    composition = PyComposition()
    for elem, isotope, number in atom_pattern.findall(formula):
        composition[_make_isotope_string(elem, int(isotope) if isotope else 0)] += int(number)
    return composition


try:
    _has_c = True
    _parse_formula = parse_formula
    _PyComposition = PyComposition
    from ._c.composition import parse_formula, PyComposition
except ImportError as e:
    print(e)
    _has_c = False


SimpleComposition = PyComposition



from brainpy._c.composition cimport (
    _ElementTable, Element, Isotope, Composition, ElementHashTable)

from brainpy._c.double_vector cimport DoubleVector

ctypedef DoubleVector dvec


cdef struct PolynomialParameters:
    dvec* elementary_symmetric_polynomial
    dvec* power_sum


cdef struct PhiConstants:
    int order
    Element* element
    PolynomialParameters* element_coefficients
    PolynomialParameters* mass_coefficients


cdef struct IsotopicConstants:
    int order
    PhiConstants** constants
    size_t size
    size_t used

cdef size_t DEFAULT_ISOTOPIC_CONSTANTS_SIZE = 7


cdef dvec* vietes(dvec* coefficients) nogil
cdef void _update_power_sum(dvec* ps_vec, dvec* esp_vec, int order) nogil
cdef void _update_elementary_symmetric_polynomial(dvec* ps_vec, dvec* esp_vec, int order) nogil
cdef void newton(dvec* ps_vec, dvec* esp_vec, int order) nogil
cdef dvec* compute_isotopic_coefficients(Element* element, bint with_mass, dvec* accumulator) nogil
cdef PolynomialParameters* make_polynomial_parameters(Element* element, bint with_mass, dvec* accumulator) nogil
cdef void print_polynomial_parameters(PolynomialParameters* params) nogil
cdef void free_polynomial_parameters(PolynomialParameters* params) nogil


cdef void print_phi_constants(PhiConstants* constants) nogil
cdef void free_phi_constants(PhiConstants* constants) nogil


cdef IsotopicConstants* make_isotopic_constants() nogil
cdef int isotopic_constants_resize(IsotopicConstants* ics) nogil
cdef void free_isotopic_constants(IsotopicConstants* isotopes) nogil

cdef void isotopic_constants_add_element(IsotopicConstants* isotopes, char* element_symbol) nogil
cdef int isotopic_constants_get(IsotopicConstants* isotopes, char* element_symbol, PhiConstants** out) nogil
cdef void isotopic_constants_update_coefficients(IsotopicConstants* isotopes) nogil

cdef double isotopic_constants_nth_element_power_sum(IsotopicConstants* isotopes, char* symbol, int order) nogil
cdef double isotopic_constants_nth_modified_element_power_sum(IsotopicConstants* isotopes, char* symbol, int order) nogil

cdef double isotopic_constants_nth_element_power_sum_by_index(IsotopicConstants* isotopes, size_t index, int order) nogil
cdef double isotopic_constants_nth_modified_element_power_sum_by_index(IsotopicConstants* isotopes, size_t index, int order) nogil

cdef void print_isotopic_constants(IsotopicConstants* isotopes) nogil



# cython: embedsignature=True

cimport cython

from brainpy._c.composition cimport (
    Element, Isotope, Composition, ElementHashTable,
    element_max_neutron_shift, _parse_isotope_string,
    _ElementTable, element_hash_table_get, make_fixed_isotope_element,
    element_hash_table_put, make_element_hash_table, free_element_hash_table)

from brainpy._c.double_vector cimport(
    DoubleVector, make_double_vector, double_vector_append,
    free_double_vector, print_double_vector, double_vector_to_list,
    reset_double_vector, make_double_vector_with_size)

from libc.stdlib cimport malloc, free, realloc
from libc.string cimport strcmp
from libc cimport *

cdef extern from * nogil:
    int printf (const char *template, ...)

# -----------------------------------------------------------------------------
# PolynomialParameters Methods

@cython.cdivision
cdef dvec* vietes(dvec* coefficients) nogil:
    cdef:
        DoubleVector* elementary_symmetric_polynomial
        size_t i
        double tail
        int sign
        double el

    elementary_symmetric_polynomial = make_double_vector_with_size(
        coefficients.used)
    tail = coefficients.v[coefficients.used - 1]

    for i in range(coefficients.used):
        sign = 1 if (i % 2) == 0 else -1
        el = sign * coefficients.v[coefficients.used - i - 1] / tail
        double_vector_append(elementary_symmetric_polynomial, el)

    return elementary_symmetric_polynomial


cdef void _update_power_sum(dvec* ps_vec, dvec* esp_vec, int order) nogil:
    cdef:
        size_t begin, end, k, j
        int sign
        double temp_ps

    begin = ps_vec.used
    end = esp_vec.used

    for k in range(begin, end):
        if k == 0:
            double_vector_append(ps_vec, 0.0)
            continue
        temp_ps = 0.
        sign = -1

        for j in range(1, k):
            sign *= -1
            temp_ps += sign * (esp_vec.v[j]) * (ps_vec.v[k - j])
        sign *= -1
        temp_ps += sign * (esp_vec.v[k]) * k
        double_vector_append(ps_vec, temp_ps)


@cython.cdivision
cdef void _update_elementary_symmetric_polynomial(dvec* ps_vec, dvec* esp_vec, int order) nogil:
    cdef:
        size_t begin, end, k, j
        int sign
        double el

    begin = esp_vec.used
    end = ps_vec.used

    for k in range(begin, end):
        if k == 0:
            double_vector_append(esp_vec, 1.0)
        elif k > order:
            double_vector_append(esp_vec, 0.0)
        else:
            el = 0.
            for j in range(1, k + 1):
                sign = 1 if (j % 2) == 1 else -1
                el += sign * ps_vec.v[j] * esp_vec.v[k - j]
            el /= <double>k
            double_vector_append(esp_vec, el)


cdef void newton(dvec* ps_vec, dvec* esp_vec, int order) nogil:
    if ps_vec.used > esp_vec.used:
        _update_elementary_symmetric_polynomial(ps_vec, esp_vec, order)
    elif ps_vec.used < esp_vec.used:
        _update_power_sum(ps_vec, esp_vec, order)


cdef dvec* compute_isotopic_coefficients(Element* element, bint with_mass, dvec* accumulator) nogil:
    cdef:
        int max_isotope_number, current_order
        Isotope* isotope
        double coef
        size_t i, j, k

    max_isotope_number = element_max_neutron_shift(element)
    for i in range(element.isotopes.size):
        k = element.isotopes.size - i - 1
        isotope = &(element.isotopes.bins[k])
        current_order = max_isotope_number - isotope.neutron_shift
        if with_mass:
            coef = isotope.mass
        else:
            coef = 1.0
        if current_order > accumulator.used:
            for j in range(accumulator.used, current_order):
                double_vector_append(accumulator, 0.)
            double_vector_append(accumulator, isotope.abundance * coef)
        elif current_order == accumulator.used:
            double_vector_append(accumulator, isotope.abundance * coef)
        else:
            printf("Error, unordered isotopes for %s\n", element.symbol)
    return accumulator

cdef PolynomialParameters* make_polynomial_parameters(Element* element, bint with_mass, dvec* accumulator) nogil:
    cdef:
        dvec* elementary_symmetric_polynomial
        dvec* power_sum
        PolynomialParameters* result

    compute_isotopic_coefficients(element, with_mass, accumulator)

    elementary_symmetric_polynomial = vietes(accumulator)
    power_sum = make_double_vector_with_size(elementary_symmetric_polynomial.used + 4)
    newton(power_sum, elementary_symmetric_polynomial, accumulator.used - 1)
    result = <PolynomialParameters*>malloc(sizeof(PolynomialParameters))
    result.elementary_symmetric_polynomial = elementary_symmetric_polynomial
    result.power_sum = power_sum
    return result


cdef void print_polynomial_parameters(PolynomialParameters* params) nogil:
    printf("PolynomialParameters: %d\n", <int>params)
    printf("  ")
    print_double_vector(params.elementary_symmetric_polynomial)
    printf("  ")
    print_double_vector(params.power_sum)
    printf("\n")


cdef void free_polynomial_parameters(PolynomialParameters* params) nogil:
    free_double_vector(params.elementary_symmetric_polynomial)
    free_double_vector(params.power_sum)
    free(params)

# -----------------------------------------------------------------------------
# PhiConstants Methods


cdef void print_phi_constants(PhiConstants* constants) nogil:
    printf("PhiConstants: %d\n", <int>constants)
    printf("Element: %s, Order: %d\n", constants.element.symbol, constants.order)
    printf("Element Coefficients:\n")
    print_polynomial_parameters(constants.element_coefficients)
    printf("Mass Coefficients:\n")
    print_polynomial_parameters(constants.mass_coefficients)
    printf("\n")


cdef void free_phi_constants(PhiConstants* constants) nogil:
    free_polynomial_parameters(constants.element_coefficients)
    free_polynomial_parameters(constants.mass_coefficients)
    free(constants)


# -----------------------------------------------------------------------------
# IsotopicConstants Methods

cdef size_t DEFAULT_ISOTOPIC_CONSTANTS_SIZE = 7


cdef IsotopicConstants* make_isotopic_constants() nogil:
    cdef:
        IsotopicConstants* result
    result = <IsotopicConstants*>malloc(sizeof(IsotopicConstants))
    result.constants = <PhiConstants**>malloc(sizeof(PhiConstants*) * DEFAULT_ISOTOPIC_CONSTANTS_SIZE)
    result.size = DEFAULT_ISOTOPIC_CONSTANTS_SIZE
    result.used = 0
    return result


cdef int isotopic_constants_resize(IsotopicConstants* ics) nogil:
    ics.constants = <PhiConstants**>realloc(ics.constants, sizeof(PhiConstants*) * ics.size * 2)
    ics.size *= 2
    if ics.constants == NULL:
        return -1
    return 0


cdef void free_isotopic_constants(IsotopicConstants* isotopes) nogil:
    cdef:
        size_t i

    i = 0
    while i < isotopes.used:
        free_phi_constants(isotopes.constants[i])
        i += 1
    free(isotopes.constants)
    free(isotopes)


cdef void isotopic_constants_add_element(IsotopicConstants* isotopes, char* element_symbol) nogil:
    cdef:
        Element* element
        dvec* accumulator
        int order, status, isotope_number
        PolynomialParameters* element_parameters
        PolynomialParameters* mass_parameters
        PhiConstants* phi_constants
        char* element_buffer

    status = isotopic_constants_get(isotopes, element_symbol, &phi_constants)
    if status == 0:
        return

    phi_constants = NULL

    status = element_hash_table_get(_ElementTable, element_symbol, &element)
    if status == -1:
        printf("Could not resolve element_symbol %s\n", element_symbol)
        return
    accumulator = make_double_vector()
    order = element_max_neutron_shift(element)
    element_parameters = make_polynomial_parameters(element, False, accumulator)
    reset_double_vector(accumulator)
    mass_parameters = make_polynomial_parameters(element, True, accumulator)
    free_double_vector(accumulator)
    phi_constants = <PhiConstants*>malloc(sizeof(PhiConstants))
    phi_constants.order = order
    phi_constants.element = element
    phi_constants.element_coefficients = element_parameters
    phi_constants.mass_coefficients = mass_parameters
    if isotopes.used + 1 == isotopes.size:
        isotopic_constants_resize(isotopes)
    isotopes.constants[isotopes.used] = phi_constants
    isotopes.used += 1


cdef int isotopic_constants_get(IsotopicConstants* isotopes, char* element_symbol, PhiConstants** out) nogil:
    cdef:
        size_t i
        int status

    i = 0
    while i < isotopes.used:
        if strcmp(isotopes.constants[i].element.symbol, element_symbol) == 0:
            out[0] = (isotopes.constants[i])
            return 0
        i += 1
    return 1


cdef int isotopic_constants_get_by_index(IsotopicConstants* isotopes, size_t index, PhiConstants** out) nogil:
    if index > isotopes.used:
        return 1
    else:
        out[0] = isotopes.constants[index]
        return 0


cdef void isotopic_constants_update_coefficients(IsotopicConstants* isotopes) nogil:
    cdef:
        size_t i, j
        PhiConstants* constants

    for i in range(isotopes.used):
        constants = isotopes.constants[i]

        if isotopes.order < constants.order:
            continue

        for j in range(constants.order, isotopes.order + 1):
            double_vector_append(constants.element_coefficients.elementary_symmetric_polynomial, 0.)
            double_vector_append(constants.mass_coefficients.elementary_symmetric_polynomial, 0.)

        constants.order = constants.element_coefficients.elementary_symmetric_polynomial.used

        newton(constants.element_coefficients.power_sum,
               constants.element_coefficients.elementary_symmetric_polynomial,
               constants.order)
        newton(constants.mass_coefficients.power_sum,
               constants.mass_coefficients.elementary_symmetric_polynomial,
               constants.order)


cdef double isotopic_constants_nth_element_power_sum(IsotopicConstants* isotopes, char* symbol, int order) nogil:
    cdef:
        PhiConstants* constants
    isotopic_constants_get(isotopes, symbol, &constants)
    return constants.element_coefficients.power_sum.v[order]


cdef double isotopic_constants_nth_element_power_sum_by_index(IsotopicConstants* isotopes, size_t index, int order) nogil:
    cdef:
        PhiConstants* constants
    isotopic_constants_get_by_index(isotopes, index, &constants)
    return constants.element_coefficients.power_sum.v[order]


cdef double isotopic_constants_nth_modified_element_power_sum(IsotopicConstants* isotopes, char* symbol, int order) nogil:
    cdef:
        PhiConstants* constants
    isotopic_constants_get(isotopes, symbol, &constants)
    return constants.mass_coefficients.power_sum.v[order]


cdef double isotopic_constants_nth_modified_element_power_sum_by_index(IsotopicConstants* isotopes, size_t index, int order) nogil:
    cdef:
        PhiConstants* constants
    isotopic_constants_get_by_index(isotopes, index, &constants)
    return constants.mass_coefficients.power_sum.v[order]


cdef void print_isotopic_constants(IsotopicConstants* isotopes) nogil:
    cdef:
        size_t i

    i = 0
    for i in range(isotopes.used):
        printf("%d\n", i)
        print_phi_constants(isotopes.constants[i])


def main():
    cdef:
        IsotopicConstants* ic
        dvec* coefficients
        dvec* elementary_symmetric_polynomial
        PhiConstants* constant
        Element* elem
        char* sym

    sym = "O"
    ic = make_isotopic_constants()
    print(ic.used)
    isotopic_constants_add_element(ic, sym)
    print(ic.used)
    isotopic_constants_add_element(ic, "C")
    isotopic_constants_add_element(ic, "H")

    if isotopic_constants_get(ic, "O", &constant) == 0:
        print_phi_constants(constant)
    else:
        print("Nope")

    free_isotopic_constants(ic)



# cython: embedsignature=True

cimport cython

from brainpy._c.composition cimport (
    Element, Isotope, Composition, ElementHashTable,
    mass_charge_ratio, PROTON, element_max_neutron_shift,
    composition_get_element_count, element_monoisotopic_mass,
    composition_mass, get_isotope_by_neutron_shift, dict_to_composition,
    print_composition, free_composition, count_type,
    make_element_hash_table, free_element_hash_table,
    _ElementTable, element_hash_table_get, make_fixed_isotope_element,
    _parse_isotope_string, element_hash_table_put)

from brainpy._c.double_vector cimport(
    DoubleVector, make_double_vector, double_vector_append,
    make_double_vector_with_size,
    free_double_vector, print_double_vector, reset_double_vector)

from libc.stdlib cimport malloc, free, realloc
from libc.string cimport strcmp
from libc.math cimport log, exp, sqrt
from libc cimport *

from brainpy._c.isotopic_constants cimport (
    IsotopicConstants, isotopic_constants_get, make_isotopic_constants,
    isotopic_constants_resize, free_isotopic_constants, isotopic_constants_add_element,
    isotopic_constants_update_coefficients,
    isotopic_constants_nth_element_power_sum, print_isotopic_constants,
    isotopic_constants_nth_element_power_sum_by_index,
    isotopic_constants_nth_modified_element_power_sum,
    newton)


# from double_vector cimport (
#     DoubleVector, free_double_vector, make_double_vector,
#     double_vector_append, print_double_vector)

ctypedef DoubleVector dvec


cdef extern from * nogil:
    int printf (const char *template, ...)
    void qsort (void *base, unsigned short n, unsigned short w, int (*cmp_func)(void*, void*))


# -----------------------------------------------------------------------------
# ElementPolynomialMap Declaration and Methods

cdef struct ElementPolynomialMap:
    char** elements
    dvec** polynomials
    size_t used
    size_t size

cdef ElementPolynomialMap* make_element_polynomial_map(size_t sizehint) nogil:
    cdef ElementPolynomialMap* result
    result = <ElementPolynomialMap*>malloc(sizeof(ElementPolynomialMap))
    result.elements = <char**>malloc(sizeof(char*) * sizehint)
    result.polynomials = <dvec**>malloc(sizeof(dvec*) * sizehint)
    result.size = sizehint
    result.used = 0

    return result

cdef int element_polynomial_map_set(ElementPolynomialMap* ep_map, char* element, dvec* polynomial) nogil:
    cdef:
        size_t i
        int status
        bint done

    done = False
    i = 0
    while (i < ep_map.used):
        if strcmp(element, ep_map.elements[i]) == 0:
            done = True
            ep_map.polynomials[i] = polynomial
            return 0
        i += 1
    if not done:
        ep_map.used += 1
        if ep_map.used >= ep_map.size:
            printf("Overloaded ElementPolynomialMap\n %d, %d\n", i, ep_map.size)
            return -1
        ep_map.elements[i] = element
        ep_map.polynomials[i] = polynomial
        return 0
    return 1

cdef int element_polynomial_map_get(ElementPolynomialMap* ep_map, char* element, dvec** polynomial) nogil:
    cdef:
        size_t i
        int status
        bint done

    done = False
    i = 0
    while i < ep_map.used:
        if strcmp(ep_map.elements[i], element) == 0:
            polynomial[0] = ep_map.polynomials[i]
            return 0
        i += 1
    return 1

cdef void free_element_polynomial_map(ElementPolynomialMap* ep_map) nogil:
    cdef:
        size_t i
    i = 0
    while i < ep_map.used:
        free_double_vector(ep_map.polynomials[i])
        i += 1
    free(ep_map.elements)
    free(ep_map.polynomials)
    free(ep_map)


# -----------------------------------------------------------------------------
# Peak Methods

cdef void print_peak(Peak* peak) nogil:
    printf("Peak: %f, %f, %d\n", peak.mz, peak.intensity, peak.charge)


cdef Peak* make_peak(double mz, double intensity, int charge) nogil:
    cdef Peak* peak
    peak = <Peak*>malloc(sizeof(Peak))
    peak.mz = mz
    peak.intensity = intensity
    peak.charge = charge
    return peak

# -----------------------------------------------------------------------------
# PeakList Methods

cdef PeakList* make_peak_list() nogil:
    cdef PeakList* result

    result = <PeakList*>malloc(sizeof(PeakList))
    result.peaks = <Peak*>malloc(sizeof(Peak) * 10)
    result.size = 10
    result.used = 0

    return result

cdef void free_peak_list(PeakList* peaklist) nogil:
    free(peaklist.peaks)
    free(peaklist)

cdef int resize_peak_list(PeakList* peaklist) nogil:
    cdef:
        Peak* peaks
    peaks = <Peak*>realloc(peaklist.peaks, sizeof(Peak) * peaklist.size * 2)
    if peaks == NULL:
        printf("realloc peaklist returned NULL\n")
        return -1
    peaklist.peaks = peaks
    peaklist.size *= 2
    return 0

cdef void peak_list_append(PeakList* peaklist, Peak* peak) nogil:
    if peaklist.used == peaklist.size:
        resize_peak_list(peaklist)
    peaklist.peaks[peaklist.used] = peak[0]
    peaklist.used += 1

cdef void peak_list_reset(PeakList* peaklist) nogil:
    peaklist.used = 0

@cython.cdivision
cdef PeakList* peak_list_ignore_below(PeakList* peaklist, double ignore_below, PeakList* result) nogil:
    cdef:
        double total
        PeakList* kept_tid
        size_t i, n
        Peak p

    total = 0
    n = peaklist.used

    if result == NULL:
        result = make_peak_list()

    for i in range(n):
        p = result.peaks[i]
        if (p.intensity < ignore_below) and (i > 1):
            continue
        else:
            total += p.intensity
        peak_list_append(result, &p)
    n = result.used
    for i in range(n):
        result.peaks[i].intensity /= total
    return result

@cython.cdivision
cdef PeakList* peak_list_truncate_after(PeakList* peaklist, double truncate_after, PeakList* result) nogil:
    cdef:
        double cumsum
        Peak peak
        size_t i, n

    cumsum = 0
    n = peaklist.used
    if result == NULL:
        result = make_peak_list()

    for i in range(n):
        peak = peaklist.peaks[i]
        cumsum += peak.intensity
        peak_list_append(result, &peak)
        if cumsum >= truncate_after:
            break

    n = result.used
    for i in range(n):
        result.peaks[i].intensity /= cumsum
    return result

cdef void peak_list_shift(PeakList* peaklist, double shift) nogil:
    cdef:
        size_t i, n
        double delta

    n = peaklist.used
    if n == 0:
        return
    delta = shift - peaklist.peaks[0].mz
    for i in range(n):
        peaklist.peaks[i].mz += delta

# -----------------------------------------------------------------------------
# ElementCache Methods

cdef ElementCache* make_element_cache(ElementHashTable* source) nogil:
    cdef:
        ElementCache* cache
    cache = <ElementCache*>malloc(sizeof(ElementCache))
    cache.source = source
    cache.elements = <Element**>malloc(sizeof(Element*) * 10)
    cache.used = 0
    cache.size = 10
    return cache


cdef void free_element_cache(ElementCache* cache) nogil:
    free(cache.elements)
    free(cache)


cdef int resize_element_cache(ElementCache* cache) nogil:
    cdef:
        Element** values
        size_t new_size
    new_size = cache.size * 10
    values = <Element**>realloc(cache.elements, sizeof(Element*) * new_size)
    if values == NULL:
        printf("resize_element_cache returned -1\n")
        return -1
    cache.elements = values
    cache.size = new_size
    return 0


cdef int element_cache_put(ElementCache* cache, Element** element) nogil:
    cdef:
        size_t i
    if (cache.used + 1) == cache.size:
        resize_element_cache(cache)
    cache.elements[cache.used] = element[0]
    cache.used += 1
    return 0


cdef int element_cache_get(ElementCache* cache, char* symbol, Element** out) nogil:
    cdef:
        size_t i
        Element* element

    if cache == NULL:
        return element_hash_table_get(_ElementTable, symbol, out)

    for i in range(cache.used):
        element = cache.elements[i]
        if strcmp(element.symbol, symbol) == 0:
            out[0] = element
            return 0
    element_hash_table_get(cache.source, symbol, out)
    element_cache_put(cache, out)
    return 1


# -----------------------------------------------------------------------------
# Support Functions

cdef int max_variants(Composition* composition, ElementCache* cache) nogil:
    cdef:
        size_t i
        int max_variants
        double count
        char* symbol
        Element* element

    if composition.max_variants != 0:
        return composition.max_variants

    max_variants = 0
    for i in range(composition.used):
        symbol = composition.elements[i]
        count = composition.counts[i]
        if cache == NULL:
            element_hash_table_get(_ElementTable, symbol, &element)
        else:
            element_cache_get(cache, symbol, &element)
        max_variants += <int>(count) * element_max_neutron_shift(element)
    composition.max_variants = max_variants
    return max_variants


cdef void validate_composition(Composition* composition) nogil:
    cdef:
        size_t i
        char* element_symbol
        char* element_buffer
        int status, isotope_number
        Element* element

    for i in range(composition.used):
        element_symbol = composition.elements[i]
        status = element_hash_table_get(_ElementTable, element_symbol, &element)
        # printf("Element %s, Status %d\n", element_symbol, status)
        if status == -1:
            element_buffer = <char*>malloc(sizeof(char) * 10)
            # printf("Could not resolve element %s, attempting to generate fixed isotope\n", element_symbol)
            _parse_isotope_string(element_symbol, &isotope_number, element_buffer)
            # printf("Element: %s, Isotope: %d\n", element_buffer, isotope_number)
            if isotope_number != 0:
                status = element_hash_table_get(_ElementTable, element_buffer, &element)
                # printf("Retreived Base Element: %s. Status: %d\n", element_buffer, status)
                element = make_fixed_isotope_element(element, isotope_number)
                element_hash_table_put(_ElementTable, element)
                free(element_buffer)
            else:
                # printf("Could not resolve element %s\n", element_symbol)
                free(element_buffer)
                return

# -----------------------------------------------------------------------------
# IsotopicDistribution Methods

cdef void isotopic_distribution_update_isotopic_constants(IsotopicDistribution* distribution) nogil:
    cdef:
        size_t i
        Composition* composition
        IsotopicConstants* isotopes
        char* symbol

    composition = distribution.composition
    isotopes = distribution._isotopic_constants

    for i in range(composition.used):
        symbol = composition.elements[i]
        isotopic_constants_add_element(isotopes, symbol)
    isotopes.order = distribution.order
    isotopic_constants_update_coefficients(isotopes)


cdef void isotopic_distribution_update_order(IsotopicDistribution* distribution, int order) nogil:
    cdef:
        int possible_variants

    possible_variants = max_variants(distribution.composition, NULL)
    if order == -1:
        distribution.order = possible_variants
    else:
        distribution.order = min(order, possible_variants)

    isotopic_distribution_update_isotopic_constants(distribution)


cdef Peak* isotopic_distribution_make_monoisotopic_peak(IsotopicDistribution* distribution) nogil:
    cdef:
        Peak* peak
        size_t i
        Element* element
        int status
        double intensity

    peak = <Peak*>malloc(sizeof(Peak))
    peak.mz = composition_mass(distribution.composition)

    intensity = 0
    for i in range(distribution.composition.used):
        status = element_cache_get(distribution.cache, distribution.composition.elements[i], &element)
        intensity += log(get_isotope_by_neutron_shift(element.isotopes, 0).abundance)
    intensity = exp(intensity)
    peak.intensity = intensity
    peak.charge = 0
    return peak


cdef IsotopicDistribution* make_isotopic_distribution(Composition* composition, int order, ElementCache* cache=NULL) nogil:
    cdef:
        IsotopicDistribution* result
    if cache == NULL:
        cache = make_element_cache(_ElementTable)
    result = <IsotopicDistribution*>malloc(sizeof(IsotopicDistribution))
    result.composition = composition
    result.cache = cache
    result.order = 0
    result._isotopic_constants = make_isotopic_constants()
    isotopic_distribution_update_order(result, order)
    result.average_mass = 0
    result.monoisotopic_peak = isotopic_distribution_make_monoisotopic_peak(result)
    return result


cdef void free_isotopic_distribution(IsotopicDistribution* distribution)  nogil:
    free(distribution.monoisotopic_peak)
    free_isotopic_constants(distribution._isotopic_constants)
    if distribution.cache != NULL:
        free_element_cache(distribution.cache)
    free(distribution)


cdef dvec* id_phi_values(IsotopicDistribution* distribution) nogil:
    cdef:
        dvec* power_sum
        size_t i
    power_sum = make_double_vector_with_size(distribution.order)
    double_vector_append(power_sum, 0.)
    for i in range(1, distribution.order + 1):
        double_vector_append(power_sum, _id_phi_value(distribution, i))
    return power_sum


cdef double _id_phi_value(IsotopicDistribution* distribution, int order) nogil:
    cdef:
        double phi
        char* element
        double count
        size_t i

    phi = 0
    i = 0
    while i < distribution.composition.used:
        element = distribution.composition.elements[i]
        count = distribution.composition.counts[i]
        phi += isotopic_constants_nth_element_power_sum(
            distribution._isotopic_constants, element, order) * count
        i += 1
    return phi


cdef dvec* id_modified_phi_values(IsotopicDistribution* distribution, char* element, dvec* power_sum) nogil:
    cdef:
        size_t i

    double_vector_append(power_sum, 0.)
    for i in range(1, distribution.order + 1):
        double_vector_append(power_sum,
                             _id_modified_phi_value(distribution, element, i))
    return power_sum


cdef double _id_modified_phi_value(IsotopicDistribution* distribution, char* symbol, int order) nogil:
    cdef:
        double phi
        char* element
        double count
        size_t i
        double coef

    phi = 0
    i = 0
    while i < distribution.composition.used:
        element = distribution.composition.elements[i]
        count = distribution.composition.counts[i]

        if strcmp(element, symbol) == 0:
            coef = count - 1
        else:
            coef = count

        phi += isotopic_constants_nth_element_power_sum(
            distribution._isotopic_constants, element, order) * coef
        i += 1

    phi += isotopic_constants_nth_modified_element_power_sum(
        distribution._isotopic_constants, symbol, order)

    return phi


cdef dvec* id_probability_vector(IsotopicDistribution* distribution) nogil:
    cdef:
        dvec* phi_values
        int max_variant_count
        dvec* probability_vector
        size_t i
        int sign

    phi_values = id_phi_values(distribution)
    probability_vector = make_double_vector_with_size(phi_values.size)
    max_variant_count = max_variants(distribution.composition, distribution.cache)

    newton(phi_values, probability_vector, max_variant_count)

    for i in range(0, probability_vector.used):
        sign = 1 if i % 2 == 0 else -1
        probability_vector.v[i] *= distribution.monoisotopic_peak.intensity * sign

    free_double_vector(phi_values)
    return probability_vector


@cython.cdivision
cdef dvec* id_center_mass_vector(IsotopicDistribution* distribution, dvec* probability_vector) nogil:
    cdef:
        dvec* mass_vector
        dvec* power_sum
        dvec* ele_sym_poly
        int max_variant_count, sign
        Element* element_struct
        char* element
        count_type _element_count
        double center, temp, polynomial_term
        double _monoisotopic_mass, base_intensity
        size_t i, j, k
        ElementPolynomialMap* ep_map

    mass_vector = make_double_vector_with_size(probability_vector.size + 3)
    power_sum = make_double_vector()
    max_variant_count = max_variants(distribution.composition, distribution.cache)
    base_intensity = distribution.monoisotopic_peak.intensity
    ep_map = make_element_polynomial_map(distribution.composition.size)

    j = 0
    while j < distribution.composition.used:
        element = distribution.composition.elements[j]
        reset_double_vector(power_sum)
        power_sum = id_modified_phi_values(distribution, element, power_sum)
        ele_sym_poly = make_double_vector()
        newton(power_sum, ele_sym_poly, max_variant_count)
        element_polynomial_map_set(ep_map, element, ele_sym_poly)
        j += 1
    free_double_vector(power_sum)
    i = 0
    for i in range(distribution.order + 1):
        sign = 1 if i % 2 == 0 else -1
        center = 0.0
        k = 0
        while k < ep_map.used:
            element = ep_map.elements[k]
            element_polynomial_map_get(ep_map, element, &ele_sym_poly)

            composition_get_element_count(distribution.composition, element, &_element_count)
            element_cache_get(distribution.cache, element, &element_struct)

            _monoisotopic_mass = element_monoisotopic_mass(element_struct)

            polynomial_term = ele_sym_poly.v[i]
            temp = _element_count
            temp *= sign * polynomial_term
            temp *= base_intensity * _monoisotopic_mass
            center += temp
            k += 1
        if probability_vector.v[i] == 0:
            double_vector_append(mass_vector, 0)
        else:
            double_vector_append(mass_vector, center / probability_vector.v[i])

    free_element_polynomial_map(ep_map)

    return mass_vector


@cython.cdivision
cdef PeakList* id_aggregated_isotopic_variants(IsotopicDistribution* distribution, int charge, double charge_carrier) nogil:
    cdef:
        dvec* probability_vector
        dvec* center_mass_vector
        PeakList* peak_set
        double average_mass, adjusted_mz
        double total
        size_t i
        Peak peak
        double center_mass_i, intensity_i

    probability_vector = id_probability_vector(distribution)
    center_mass_vector = id_center_mass_vector(distribution, probability_vector)

    peak_set = make_peak_list()

    average_mass = 0
    total = 0

    for i in range(probability_vector.used):
        total += probability_vector.v[i]

    for i in range(distribution.order + 1):
        center_mass_i = center_mass_vector.v[i]
        intensity_i = probability_vector.v[i]


        if charge != 0:
            adjusted_mz = mass_charge_ratio(center_mass_i, charge, charge_carrier)
        else:
            adjusted_mz = center_mass_i


        peak.mz = adjusted_mz
        peak.intensity = intensity_i / total
        peak.charge = charge

        if peak.intensity < 1e-10:
            continue

        peak_list_append(peak_set, &peak)

        average_mass += adjusted_mz * intensity_i

    average_mass /= total

    free_double_vector(probability_vector)
    free_double_vector(center_mass_vector)

    distribution.average_mass = average_mass

    sort_by_mz(peak_set)
    return peak_set

cdef void sort_by_mz(PeakList* peaklist) nogil:
    qsort(peaklist.peaks, peaklist.used, sizeof(Peak), compare_by_mz)

cdef int compare_by_mz(const void * a, const void * b) nogil:
    if (<Peak*>a).mz < (<Peak*>b).mz:
        return -1
    elif (<Peak*>a).mz == (<Peak*>b).mz:
        return 0
    elif (<Peak*>a).mz > (<Peak*>b).mz:
        return 1


cpdef bint check_composition_non_negative(dict composition):
    cdef:
        bint negative_element
        str k
        object v

    negative_element = False
    for k, v in composition.items():
        if v < 0:
            negative_element = True
            break
    return negative_element


cdef int guess_npeaks(Composition* composition_struct, size_t max_npeaks, ElementCache* cache=NULL) nogil:
    cdef:
        int max_n_variants, npeaks
    max_n_variants = max_variants(composition_struct, cache)
    npeaks = <int>sqrt(max_n_variants) - 2
    npeaks = min(max(npeaks, 3), max_npeaks)
    return npeaks


def pyisotopic_variants(object composition not None, object npeaks=None, int charge=0,
                        double charge_carrier=PROTON):
    '''
    Compute a peak list representing the theoretical isotopic cluster for `composition`.

    Parameters
    ----------
    composition : Mapping
        Any Mapping type where keys are element symbols and values are integers
    n_peaks: int
        The number of peaks to include in the isotopic cluster, starting from the monoisotopic peak.
        If given a number below 1 or above the maximum number of isotopic variants, the maximum will
        be used. If `None`, a "reasonable" value is chosen by `int(sqrt(max_variants(composition)))`.
    charge: int
        The charge state of the isotopic cluster to produce. Defaults to 0, theoretical neutral mass.
    charge_carrier: double
        The mass of the molecule contributing the ion's charge

    Returns
    -------
    list of TheoreticalPeak
    '''
    return _isotopic_variants(composition, npeaks, charge, charge_carrier)


cpdef list _isotopic_variants(object composition, object npeaks=None, int charge=0, double charge_carrier=PROTON):
    '''
    Compute a peak list representing the theoretical isotopic cluster for `composition`.

    Parameters
    ----------
    composition : Mapping
        Any Mapping type where keys are element symbols and values are integers
    n_peaks: int
        The number of peaks to include in the isotopic cluster, starting from the monoisotopic peak.
        If given a number below 1 or above the maximum number of isotopic variants, the maximum will
        be used. If `None`, a "reasonable" value is chosen by `int(sqrt(max_variants(composition)))`.
    charge: int
        The charge state of the isotopic cluster to produce. Defaults to 0, theoretical neutral mass.
    charge_carrier: double
        The mass of the molecule contributing the ion's charge

    Returns
    -------
    list of TheoreticalPeak
    '''
    cdef:
        Composition* composition_struct
        list peaklist
        PeakList* peak_set
        IsotopicDistribution* dist
        int _npeaks, max_n_variants

    composition_struct = dict_to_composition(dict(composition))
    validate_composition(composition_struct)

    if npeaks is None:
        _npeaks = guess_npeaks(composition_struct, 300)
    else:
        # The npeaks variable is left as a Python-level variable to
        # allow it to be any Python numeric type
        _npeaks = npeaks - 1

    dist = make_isotopic_distribution(composition_struct, _npeaks)
    peak_set = id_aggregated_isotopic_variants(dist, charge, charge_carrier)

    peaklist = peaklist_to_list(peak_set)

    free_peak_list(peak_set)
    free_isotopic_distribution(dist)
    free_composition(composition_struct)
    return peaklist


cdef PeakList* isotopic_variants(Composition* composition, int npeaks, int charge=0, double charge_carrier=PROTON) nogil:
    cdef:
        IsotopicDistribution* dist
        PeakList* peaklist
        int max_n_variants

    validate_composition(composition)

    if npeaks == 0:
        npeaks = guess_npeaks(composition, 300)
    else:
        npeaks = npeaks - 1

    dist = make_isotopic_distribution(composition, npeaks)
    peaklist = id_aggregated_isotopic_variants(dist, charge, charge_carrier)
    free_isotopic_distribution(dist)

    return peaklist


cdef list peaklist_to_list(PeakList* peaklist):
    cdef:
        list pypeaklist
        size_t i
    pypeaklist = []
    for i in range(peaklist.used):
        pypeaklist.append(TheoreticalPeak._create(
            peaklist.peaks[i].mz, peaklist.peaks[i].intensity, peaklist.peaks[i].charge))
    return pypeaklist


@cython.freelist(1000000)
cdef class TheoreticalPeak(object):
    def __init__(self, mz, intensity, charge):
        self.mz = mz
        self.intensity = intensity
        self.charge = charge

    def __repr__(self):
        return "Peak(mz=%f, intensity=%f, charge=%d)" % (self.mz, self.intensity, self.charge)

    cpdef bint _eq(self, TheoreticalPeak other):
        cdef bint val
        val = (abs(self.mz - other.mz) < 1e-10 and\
               abs(self.intensity - other.intensity) < 1e-10 and\
               self.charge == other.charge)
        return val

    def __hash__(self):
        return hash((self.mz, self.intensity, self.charge))

    def __richcmp__(self, other, int code):
        if code == 2:
            return self._eq(other)
        elif code == 3:
            return not self._eq(other)

    def __reduce__(self):
        return TheoreticalPeak, (self.mz, self.intensity, self.charge)

    cpdef TheoreticalPeak clone(self):
        return TheoreticalPeak._create(self.mz, self.intensity, self.charge)

    @staticmethod
    cdef TheoreticalPeak _create(double mz, double intensity, int charge):
        cdef:
            TheoreticalPeak inst
        inst = TheoreticalPeak.__new__(TheoreticalPeak)
        inst.mz = mz
        inst.intensity = intensity
        inst.charge = charge
        return inst


def main():
    cdef:
        dict comp_dict
        Composition* composition
        IsotopicDistribution* distribution
        IsotopicDistribution* distribution2

    comp_dict = dict(H=2, O=1)
    print comp_dict
    composition = dict_to_composition(comp_dict)
    print_composition(composition)
    distribution = make_isotopic_distribution(composition, 4)
    print "Going to print constants"
    print_isotopic_constants(distribution._isotopic_constants)
    print "Done"

    print "Trying to free"
    free_isotopic_distribution(distribution)
    print "Free Done"

    distribution2 = make_isotopic_distribution(composition, 4)
    print "Seconc construction"
    print_isotopic_constants(distribution2._isotopic_constants)
    print "Second Free"
    free_isotopic_distribution(distribution2)
    print comp_dict
    free_composition(composition)
    print "Really done"


def test(object composition, int max_npeaks=300):
    cdef:
        Composition* composition_struct
        IsotopicDistribution* distribution
        int npeaks, max_n_variants

    composition_struct = dict_to_composition(dict(composition))
    validate_composition(composition_struct)

    npeaks = guess_npeaks(composition_struct, max_npeaks)
    print("Guessed # of Peaks: ", npeaks)

    dist = make_isotopic_distribution(composition_struct, npeaks)

    print("Size of probability vector:", id_probability_vector(dist).used)

    peak_set = id_aggregated_isotopic_variants(dist, 1, PROTON)

    peaklist = peaklist_to_list(peak_set)

    print("Actual # of Peaks Returned:", len(peaklist))

    free_peak_list(peak_set)
    free_isotopic_distribution(dist)
    free_composition(composition_struct)
    return peaklist



import unittest

from brainpy import IsotopicDistribution, isotopic_variants, calculate_mass, neutral_mass, _has_c, Peak

if _has_c:
    from brainpy.brainpy import _IsotopicDistribution


class TestIsotopicDistribution(unittest.TestCase):
    def test_neutral_mass(self):
        hexnac = {'H': 13, 'C': 8, 'O': 5, 'N': 1}
        dist = isotopic_variants(hexnac)

        reference = [
            Peak(mz=203.079373, intensity=0.901867, charge=0),
            Peak(mz=204.082545, intensity=0.084396, charge=0),
            Peak(mz=205.084190, intensity=0.012787, charge=0),
            Peak(mz=206.086971, intensity=0.000950, charge=0)
        ]
        for inst, ref in zip(dist, reference):
            self.assertAlmostEqual(inst.mz, ref.mz, 3)
            self.assertAlmostEqual(inst.intensity, ref.intensity, 3)

    if _has_c:
        def test_pure_python(self):
            hexnac = {'H': 13, 'C': 8, 'O': 5, 'N': 1}
            dist = _IsotopicDistribution(hexnac, 4).aggregated_isotopic_variants()

            reference = [
                Peak(mz=203.079373, intensity=0.901867, charge=0),
                Peak(mz=204.082545, intensity=0.084396, charge=0),
                Peak(mz=205.084190, intensity=0.012787, charge=0),
                Peak(mz=206.086971, intensity=0.000950, charge=0)
            ]
            for inst, ref in zip(dist, reference):
                self.assertAlmostEqual(inst.mz, ref.mz, 3)
                self.assertAlmostEqual(inst.intensity, ref.intensity, 3)


if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import

import operator
import re

from collections import OrderedDict, Counter
from math import exp, log, sqrt
from sys import float_info

from brainpy.mass_dict import nist_mass
from brainpy.composition import (
    PyComposition,
    parse_formula,
    calculate_mass,
    _make_isotope_string,
    _get_isotope)

mz_getter = operator.attrgetter("mz")
PROTON = nist_mass["H+"][0][0]
MACHINE_EPSILON = float_info.epsilon
ZERO = 0.
ONE = 1.0


def neutral_mass(mz, z, charge_carrier=PROTON):
    return (mz * abs(z)) - (z * charge_carrier)


def mass_charge_ratio(neutral_mass, z, charge_carrier=PROTON):
    return (neutral_mass + (z * charge_carrier)) / abs(z)


def give_repr(cls):  # pragma: no cover
    r"""Patch a class to give it a generic __repr__ method
    that works by inspecting the instance dictionary.

    Parameters
    ----------
    cls: type
        The class to add a generic __repr__ to.

    Returns
    -------
    cls: type
        The passed class is returned
    """
    def reprer(self):
        attribs = ', '.join(["%s=%r" % (k, v) for k, v in self.__dict__.items() if not k.startswith("_")])
        wrap = "{self.__class__.__name__}({attribs})".format(self=self, attribs=attribs)
        return wrap
    cls.__repr__ = reprer
    return cls


@give_repr
class PolynomialParameters(object):
    def __init__(self, elementary_symmetric_polynomial, power_sum):
        self.elementary_symmetric_polynomial = elementary_symmetric_polynomial
        self.power_sum = power_sum

    def __iter__(self):
        yield self.power_sum
        yield self.elementary_symmetric_polynomial


@give_repr
class PhiConstants(object):
    def __init__(self, order, element, element_coefficients, mass_coefficients):
        self.order = order
        self.element = element
        self.element_coefficients = element_coefficients
        self.mass_coefficients = mass_coefficients


def newton(power_sum, elementary_symmetric_polynomial, order):
    r'''
    Given two lists of values, the first list being the `power sum`s of a
    polynomial, and the second being expressions of the roots of the
    polynomial as found by Viete's Formula, use information from the longer list to
    fill out the shorter list using Newton's Identities.

    .. note::
        Updates are done **in place**

    Parameters
    ----------
    power_sum: list of float
    elementary_symmetric_polynomial: list of float
    order: int
        The number of terms to expand to when updating `elementary_symmetric_polynomial`

    See Also
    --------
    https://en.wikipedia.org/wiki/Newton%27s_identities[https://en.wikipedia.org/wiki/Newton%27s_identities]
    '''
    if len(power_sum) > len(elementary_symmetric_polynomial):
        _update_elementary_symmetric_polynomial(power_sum, elementary_symmetric_polynomial, order)
    elif len(power_sum) < len(elementary_symmetric_polynomial):
        _update_power_sum(power_sum, elementary_symmetric_polynomial, order)


def _update_elementary_symmetric_polynomial(power_sum, elementary_symmetric_polynomial, order):
    begin = len(elementary_symmetric_polynomial)
    end = len(power_sum)
    for k in range(begin, end):
        if k == 0:
            elementary_symmetric_polynomial.append(1.0)
        elif k > order:
            elementary_symmetric_polynomial.append(0.)
        else:
            el = 0.
            for j in range(1, k + 1):
                sign = 1 if (j % 2) == 1 else -1
                el += sign * power_sum[j] * elementary_symmetric_polynomial[k - j]
            el /= float(k)
            elementary_symmetric_polynomial.append(el)


def _update_power_sum(ps_vec, esp_vec, order):
    begin = len(ps_vec)
    end = len(esp_vec)
    for k in range(begin, end):
        if k == 0:
            ps_vec.append(0.)
            continue
        temp_ps = 0.
        sign = -1
        for j in range(1, k):
            sign *= -1
            temp_ps += sign * esp_vec[j] * ps_vec[k - j]
        sign *= -1
        temp_ps += sign * esp_vec[k] * k
        ps_vec.append(temp_ps)


def vietes(coefficients):
    r'''
    Given the coefficients of a polynomial of a single variable,
    compute an elementary symmetric polynomial of the roots of the
    input polynomial by Viete's Formula:

    .. math::
        \sum_{1\le i_1<i_2<...<i_k \le n} x_{i_1}x_{i_2}...x_{i_k} = (-1)^k\frac{a_{n-k}}{a_n}

    Parameters
    ----------
    coefficients: list of float
        A list of coefficients of the input polynomial of arbitrary length (degree) `n`

    Returns
    -------
    list of float:
        A list containing the expression of the roots of the input polynomial of length `n`

    See Also
    --------
    https://en.wikipedia.org/wiki/Vieta%27s_formulas

    '''
    elementary_symmetric_polynomial = []
    tail = float(coefficients[-1])
    size = len(coefficients)

    for i in range(size):
        sign = 1 if (i % 2) == 0 else -1
        el = sign * coefficients[size - i - 1] / tail
        elementary_symmetric_polynomial.append(el)
    return elementary_symmetric_polynomial


@give_repr
class Isotope(object):
    '''
    Isotope represents an elenent with an integer number of neutrons specified.

    Attributes
    ----------
    mass: float
        Measurable mass in Da
    abundance: float [0.0:1.0]
        The abundance of this isotope in nature. This number is between 0.0 and 1.0
    neutron_shift: int
        The number of neutrons different between this isotope and the "normal" form. May be 0
        if this represents that normal form.
    '''
    def __init__(self, mass, abundance, neutron_shift, neutrons):
        self.mass = mass
        self.abundance = abundance
        self.neutrons = neutrons
        self.neutron_shift = neutron_shift


@give_repr
class Element(object):
    def __init__(self, symbol, isotopes=None):
        self.symbol = symbol
        self.isotopes = isotopes or _isotopes_of(symbol)

    def average_mass(self):  # pragma: no cover
        mass = 0.0
        for i, isotope in self.isotopes.items():
            mass += isotope.mass * isotope.abundance
        return mass

    def monoisotopic_mass(self):
        return self.isotopes[0].mass

    def __iter__(self):
        return iter(self.isotopes.values())

    def max_neutron_shift(self):
        if len(self.isotopes) == 0:
            return 0
        return list(self.isotopes.values())[-1].neutron_shift

    def min_neutron_shift(self):
        if len(self.isotopes) == 0:
            return -1
        return list(self.isotopes.values())[0].neutron_shift

    def __eq__(self, other):
        if other is None:
            return False
        if isinstance(other, str):
            return self.symbol == other
        else:
            return self.symbol == other.symbol

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.symbol)


def make_fixed_isotope_element(element, neutrons):
    isotope = element.isotopes[neutrons - element.isotopes[0].neutrons]
    el = Element(element.symbol + ("[%d]" % neutrons), OrderedDict([
            (0, Isotope(isotope.mass, abundance=1.0, neutron_shift=0, neutrons=neutrons)),
        ]))
    return el


def _isotopes_of(element):
    freqs = dict()
    for i, mass_freqs in nist_mass[element].items():
        if i == 0:
            continue
        if mass_freqs[1] > 0:
            freqs[i] = mass_freqs
    if len(freqs) == 0:
        return OrderedDict()
    mono_neutrons = max(freqs.items(), key=lambda x: x[1][1])[0]
    freqs = OrderedDict(sorted(((k - mono_neutrons, Isotope(*v, neutron_shift=k - mono_neutrons, neutrons=k))
                                for k, v in freqs.items()), key=lambda x: x[0]))
    return freqs


periodic_table = {k: Element(k) for k in nist_mass}
# periodic_table = {k: e for k, e in periodic_table.items() if e.max_neutron_shift() != 0 and e.min_neutron_shift() >= 0}
periodic_table["H+"] = Element("H+")


class IsotopicConstants(dict):
    def __init__(self, order):
        self._order = 0
        self.order = order

    @property
    def order(self):
        return self._order

    @order.setter
    def order(self, value):
        self._order = value
        self.update_coefficients()

    def coefficients(self, element, with_mass=False):
        max_isotope_number = element.max_neutron_shift()
        accumulator = []
        for isotope in reversed(list(element)):
            current_order = max_isotope_number - isotope.neutron_shift
            if with_mass:
                coef = isotope.mass
            else:
                coef = 1.

            if current_order > len(accumulator):
                for i in range(len(accumulator), current_order):
                    accumulator.append(0.)
                accumulator.append(isotope.abundance * coef)
            elif current_order == len(accumulator):
                accumulator.append(isotope.abundance * coef)
            else:
                raise Exception("The list of neutron shifts is not ordered.")

        elementary_symmetric_polynomial = vietes(accumulator)
        power_sum = []
        newton(power_sum, elementary_symmetric_polynomial, len(accumulator) - 1)
        return PolynomialParameters(elementary_symmetric_polynomial, power_sum)

    def add_element(self, symbol):
        if symbol in self:
            return
        if symbol in periodic_table:
            element = periodic_table[symbol]
        else:
            symbol_parsed, isotope = _get_isotope(symbol)
            if isotope == 0:
                raise KeyError(symbol)
            element = make_fixed_isotope_element(periodic_table[symbol_parsed], isotope)
        order = element.max_neutron_shift()
        element_parameters = self.coefficients(element)
        mass_parameters = self.coefficients(element, True)
        self[symbol] = PhiConstants(order, element, element_parameters, mass_parameters)

    def update_coefficients(self):
        for symbol, phi_constants in self.items():
            if self.order < phi_constants.order:
                continue

            for i in range(phi_constants.order, self.order + 1):
                phi_constants.element_coefficients.elementary_symmetric_polynomial.append(0.)
                phi_constants.mass_coefficients.elementary_symmetric_polynomial.append(0.)

            phi_constants.order = len(phi_constants.element_coefficients.elementary_symmetric_polynomial)
            newton(*phi_constants.element_coefficients, order=phi_constants.order)
            newton(*phi_constants.mass_coefficients, order=phi_constants.order)

    def nth_element_power_sum(self, symbol, order):
        constants = self[symbol]
        return constants.element_coefficients.power_sum[order]

    def nth_modified_element_power_sum(self, symbol, order):
        constants = self[symbol]
        return constants.mass_coefficients.power_sum[order]


def max_variants(composition):
    """Calculates the maximum number of isotopic variants that could be produced by a
    composition.

    Parameters
    ----------
    composition : Mapping
        Any Mapping type where keys are element symbols and values are integers

    Returns
    -------
        max_n_variants : int
    """
    max_n_variants = 0

    for element, count in composition.items():
        if element == "H+":
            continue
        try:
            max_n_variants += count * periodic_table[element].max_neutron_shift()
        except KeyError:
            pass

    return max_n_variants


@give_repr
class Peak(object):
    """
    Represent a single theoretical peak centroid.

    Peaks are comparable, hashable, and can be copied by calling
    :meth:`clone`

    Attributes
    ----------
    charge : int
        The charge state of the peak
    intensity : float
        The height of the peak. Peaks created as part of a
        theoretical isotopic cluster will be have an intensity
        between 0 and 1.
    mz : float
        The mass-to-charge ratio of the peak
    """
    def __init__(self, mz, intensity, charge):
        self.mz = mz
        self.intensity = intensity
        self.charge = charge

    def __eq__(self, other):  # pragma: no cover
        equal = all(
            abs(self.mz - other.mz) < 1e-10,
            abs(self.intensity - other.intensity) < 1e-10,
            self.charge == other.charge)
        return equal

    def __ne__(self, other):  # pragma: no cover
        return not (self == other)

    def __hash__(self):  # pragma: no cover
        return hash(self.mz)

    def clone(self):  # pragma: no cover
        return self.__class__(self.mz, self.intensity, self.charge)


@give_repr
class IsotopicDistribution(object):
    """
    Constructs a theoretical isotopic distribution for a given composition
    out to a given number of peaks.

    Attributes
    ----------
    average_mass : float
        The average (weighted) mass of the resulting isotopic cluster
    composition : dict
        The composition to create the isotopic cluster for
    order : int
        The number of peaks to produce and the number of terms in the
        generating polynomial expression.
    """
    def __init__(self, composition, order=-1):
        self.composition = composition
        self._isotopic_constants = IsotopicConstants(order)
        self._order = 0
        self.order = order
        self.average_mass = 0.
        self.monoisotopic_peak = self._create_monoisotopic_peak()

    @property
    def order(self):
        return self._order

    @order.setter
    def order(self, value):
        max_variant_count = max_variants(self.composition)
        if value == -1:
            self._order = max_variant_count
        else:
            self._order = min(value, max_variant_count)
        self._update_isotopic_constants()

    def _update_isotopic_constants(self):
        for element in self.composition:
            self._isotopic_constants.add_element(element)
        self._isotopic_constants.order = self._order

    def _create_monoisotopic_peak(self):
        mass = calculate_mass(self.composition)
        intensity = 0.
        for element in self.composition:
            if element == "H+":
                continue
            # intensity += log(periodic_table[element].isotopes[0].abundance)
            intensity += log(self._isotopic_constants[element].element.isotopes[0].abundance)
        intensity = exp(intensity)
        return Peak(mass, intensity, 0)

    def _phi_value(self, order):
        phi = 0.
        for element, count in self.composition.items():
            if element == "H+":
                continue
            phi += self._isotopic_constants.nth_element_power_sum(element, order) * count
        return phi

    def _modified_phi_value(self, symbol, order):
        phi = 0.
        for element, count in self.composition.items():
            if element == "H+":
                continue
            # Count is one lower for this symbol because an isotope is present
            # accounted for in the call to `nth_modified_element_power_sum` at
            # the end?
            coef = count if element != symbol else count - 1
            phi += self._isotopic_constants.nth_element_power_sum(element, order) * coef

        phi += self._isotopic_constants.nth_modified_element_power_sum(symbol, order)
        return phi

    def phi_values(self):
        power_sum = [0.]
        for i in range(1, self.order + 1):
            power_sum.append(self._phi_value(i))
        return power_sum

    def modified_phi_values(self, symbol):
        power_sum = [0.]
        for i in range(1, self.order + 1):
            power_sum.append(self._modified_phi_value(symbol, i))
        return power_sum

    def probability(self):
        phi_values = self.phi_values()
        max_variant_count = max_variants(self.composition)
        probability_vector = []
        newton(phi_values, probability_vector, max_variant_count)

        for i in range(0, len(probability_vector)):
            # The sign of each term in the probability vector (populated by
            # Newton's Identities by solving for the Elementary Symmetric Polynomial
            # given the Power Sums) alternates in the same order as `sign`.
            # This ensures that the probability vector is strictly positive.
            sign = 1 if i % 2 == 0 else -1
            # q(j) = q(0)  * e(j) * (-1)^j
            # intensity of the jth peak is |probability[j]| * the intensity of monoisotopic peak
            probability_vector[i] *= self.monoisotopic_peak.intensity * sign

        return probability_vector

    def center_mass(self, probability_vector):
        mass_vector = []
        max_variant_count = max_variants(self.composition)

        ele_sym_poly_map = dict()
        for element in self.composition:
            if element == "H+":
                continue
            power_sum = self.modified_phi_values(element)
            ele_sym_poly = []
            newton(power_sum, ele_sym_poly, max_variant_count)
            ele_sym_poly_map[element] = ele_sym_poly
        for i in range(self.order + 1):
            sign = 1 if i % 2 == 0 else -1
            center = 0.0
            for element, ele_sym_poly in ele_sym_poly_map.items():
                center += self.composition[element] * sign * ele_sym_poly[i] *\
                 self.monoisotopic_peak.intensity * self._isotopic_constants[element].element.monoisotopic_mass()
                 # self.monoisotopic_peak.intensity * periodic_table[element].monoisotopic_mass()
            mass_vector.append((center / probability_vector[i]) if probability_vector[i] > 0 else 0)
        return mass_vector

    def aggregated_isotopic_variants(self, charge=0, charge_carrier=PROTON):
        '''
        Compute the m/z (or neutral mass when `charge` == 0) for each
        aggregated isotopic peak and their intensity relative to
        the monoisotopic peak.

        Parameters
        ----------
        charge: int
            The charge state of the resulting theoretical isotopic cluster
        charge_carrier: float
            The mass added for each degree of charge

        Returns
        -------
        theoretical_isotopic_distribution: list
            A list of :class:`Peak` objects whose intensities are proportional to
            each other to reflect relative peak heights.

        '''
        probability_vector = self.probability()
        center_mass_vector = self.center_mass(probability_vector)

        peak_set = []
        average_mass = 0.
        total = sum(probability_vector)

        for i in range(self.order + 1):
            if charge != 0:
                adjusted_mz = mass_charge_ratio(center_mass_vector[i], charge, charge_carrier)
            else:
                adjusted_mz = center_mass_vector[i]
            if adjusted_mz < 1:
                continue
            peak = Peak(adjusted_mz, probability_vector[i] / total, charge)
            if peak.intensity < 0:
                continue
            peak_set.append(peak)
            average_mass += adjusted_mz * probability_vector[i]

        average_mass /= total
        self.average_mass = average_mass
        peak_set.sort(key=mz_getter)
        return tuple(peak_set)


def isotopic_variants(composition, npeaks=None, charge=0, charge_carrier=PROTON):
    '''
    Compute a peak list representing the theoretical isotopic cluster for `composition`.

    Parameters
    ----------
    composition : Mapping
        Any Mapping type where keys are element symbols and values are integers
    npeaks: int
        The number of peaks to include in the isotopic cluster, starting from the monoisotopic peak.
        If given a number below 1 or above the maximum number of isotopic variants, the maximum will
        be used. If `None`, a "reasonable" value is chosen by `int(sqrt(max_variants(composition)))`.
    charge: int
        The charge state of the isotopic cluster to produce. Defaults to 0, theoretical neutral mass.

    Returns
    -------
    list of Peaks

    See Also
    --------
    :class:`IsotopicDistribution`

    '''
    if npeaks is None:
        max_n_variants = max_variants(composition)
        npeaks = int(sqrt(max_n_variants) - 2)
        npeaks = max(npeaks, 3)
    else:
        # Monoisotopic Peak is not included
        npeaks -= 1
    return IsotopicDistribution(composition, npeaks).aggregated_isotopic_variants(
        charge, charge_carrier=charge_carrier)


try:
    _has_c = True
    _IsotopicDistribution = IsotopicDistribution
    from ._speedup import IsotopicDistribution
    from ._c.isotopic_distribution import pyisotopic_variants as isotopic_variants

except ImportError:
    _has_c = False
